from .__init__ import ctypes, SDL_SYSTEM, \
    SDL_FUNC, SDL_SET_CURRENT_BINARY, SDL_BINARY

from .SDL_init import SDL_AppResult, \
    SDL_AppInit_func, SDL_AppIterate_func, SDL_AppEvent_func, SDL_AppQuit_func

SDL_SET_CURRENT_BINARY(SDL_BINARY)

SDL_main_func = ctypes.CFUNCTYPE(ctypes.c_int, ctypes.c_int, ctypes.POINTER(ctypes.c_char_p))

SDL_FUNC("SDL_SetMainReady", None)

SDL_FUNC("SDL_RunApp", ctypes.c_int, ctypes.c_int, ctypes.POINTER(ctypes.c_char_p), SDL_main_func, ctypes.c_void_p)
SDL_FUNC("SDL_EnterAppMainCallbacks", ctypes.c_int, ctypes.c_int, ctypes.POINTER(ctypes.c_char_p), SDL_AppInit_func, SDL_AppIterate_func, SDL_AppEvent_func, SDL_AppQuit_func)

if SDL_SYSTEM in ["Windows"]:
    SDL_FUNC("SDL_RegisterApp", ctypes.c_bool, ctypes.c_char_p, ctypes.c_uint32, ctypes.c_void_p)
    SDL_FUNC("SDL_UnregisterApp", None)


import os, sys, atexit, typing

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))

import sdl3, ctypes

functions = {}

def TEST_RegisterFunction(systems: typing.List[str]) -> typing.Callable:
    return lambda func: functions.update({func: systems})

class TEST_PassFunction(Exception):
    ...

from tests.TEST_init import *
from tests.TEST_video import *

@atexit.register
def TEST_RunAllTests() -> None:
    if not functions: return
    passed, failed = 0, 0

    for func, systems in functions.items():
        try:
            sdl3.SDL_ClearError()
            if sdl3.SDL_SYSTEM in systems: func()
            else: raise TEST_PassFunction()
            print("\33[32m", f"Test '{func.__name__}' passed.", "\33[0m", sep = "", flush = True)
            passed += 1

        except TEST_PassFunction as error:
            print("\33[33m", f"Test '{func.__name__}' is not supported on {sdl3.SDL_SYSTEM}.", "\33[0m", sep = "", flush = True)

        except AssertionError as error:
            print("\33[31m", f"Test '{func.__name__}' failed: {error}", "\33[0m", sep = "", flush = True)
            failed += 1

    print("\33[35m", f"{passed} test(s) passed, {failed} test(s) failed.", "\33[0m", sep = "", flush = True)
    if failed: os._exit(-1)


import os, sys

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))

import sdl3

project = "PySDL3"
copyright = "2025, Yusuf Rençber"
author = "Yusuf Rençber"
release = sdl3.__version__

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.napoleon",
    "sphinx.ext.intersphinx",
    "sphinx.ext.viewcode",
    "sphinx.ext.todo",
    "sphinx.ext.inheritance_diagram"
]

templates_path = []
exclude_patterns = ["_build"]

language = "en"
pygments_style = None

html_theme = "sphinx_rtd_theme"
html_static_path = []


from .__init__ import ctypes, \
    SDL_FUNC, SDL_SET_CURRENT_BINARY, SDL_BINARY

from .SDL_properties import SDL_PropertiesID
from .SDL_iostream import SDL_IOStream

SDL_SET_CURRENT_BINARY(SDL_BINARY)

class SDL_Process(ctypes.c_void_p):
    ...

SDL_FUNC("SDL_CreateProcess", ctypes.POINTER(SDL_Process), ctypes.POINTER(ctypes.c_char_p), ctypes.c_bool)

SDL_ProcessIO = ctypes.c_int

SDL_PROCESS_STDIO_INHERITED = 0
SDL_PROCESS_STDIO_NULL = 1
SDL_PROCESS_STDIO_APP = 2
SDL_PROCESS_STDIO_REDIRECT = 3

SDL_FUNC("SDL_CreateProcessWithProperties", ctypes.POINTER(SDL_Process), SDL_PropertiesID)

SDL_PROP_PROCESS_CREATE_ARGS_POINTER = "SDL.process.create.args"
SDL_PROP_PROCESS_CREATE_ENVIRONMENT_POINTER = "SDL.process.create.environment"
SDL_PROP_PROCESS_CREATE_STDIN_NUMBER = "SDL.process.create.stdin_option"
SDL_PROP_PROCESS_CREATE_STDIN_POINTER = "SDL.process.create.stdin_source"
SDL_PROP_PROCESS_CREATE_STDOUT_NUMBER = "SDL.process.create.stdout_option"
SDL_PROP_PROCESS_CREATE_STDOUT_POINTER = "SDL.process.create.stdout_source"
SDL_PROP_PROCESS_CREATE_STDERR_NUMBER = "SDL.process.create.stderr_option"
SDL_PROP_PROCESS_CREATE_STDERR_POINTER = "SDL.process.create.stderr_source"
SDL_PROP_PROCESS_CREATE_STDERR_TO_STDOUT_BOOLEAN = "SDL.process.create.stderr_to_stdout"
SDL_PROP_PROCESS_CREATE_BACKGROUND_BOOLEAN = "SDL.process.create.background"

SDL_FUNC("SDL_GetProcessProperties", SDL_PropertiesID, ctypes.POINTER(SDL_Process))

SDL_PROP_PROCESS_PID_NUMBER = "SDL.process.pid"
SDL_PROP_PROCESS_STDIN_POINTER = "SDL.process.stdin"
SDL_PROP_PROCESS_STDOUT_POINTER = "SDL.process.stdout"
SDL_PROP_PROCESS_STDERR_POINTER = "SDL.process.stderr"
SDL_PROP_PROCESS_BACKGROUND_BOOLEAN = "SDL.process.background"

SDL_FUNC("SDL_ReadProcess", ctypes.c_void_p, ctypes.POINTER(SDL_Process), ctypes.POINTER(ctypes.c_size_t), ctypes.POINTER(ctypes.c_int))

SDL_FUNC("SDL_GetProcessInput", ctypes.POINTER(SDL_IOStream), ctypes.POINTER(SDL_Process))
SDL_FUNC("SDL_GetProcessOutput", ctypes.POINTER(SDL_IOStream), ctypes.POINTER(SDL_Process))

SDL_FUNC("SDL_KillProcess", ctypes.c_bool, ctypes.POINTER(SDL_Process), ctypes.c_bool)
SDL_FUNC("SDL_WaitProcess", ctypes.c_bool, ctypes.POINTER(SDL_Process), ctypes.c_bool, ctypes.POINTER(ctypes.c_int))
SDL_FUNC("SDL_DestroyProcess", None, ctypes.POINTER(SDL_Process))


from .__init__ import ctypes, \
    SDL_FUNC, SDL_SET_CURRENT_BINARY, SDL_BINARY

from .SDL_properties import SDL_PropertiesID

SDL_SET_CURRENT_BINARY(SDL_BINARY)

SDL_IOStatus = ctypes.c_uint32

SDL_IO_STATUS_READY = 0
SDL_IO_STATUS_ERROR = 1
SDL_IO_STATUS_EOF = 2
SDL_IO_STATUS_NOT_READY = 3
SDL_IO_STATUS_READONLY = 4
SDL_IO_STATUS_WRITEONLY = 5

SDL_IOWhence = ctypes.c_uint32

SDL_IO_SEEK_SET = 0
SDL_IO_SEEK_CUR = 1
SDL_IO_SEEK_END = 2

class SDL_IOStreamInterface(ctypes.Structure):
    _fields_ = [
        ("version", ctypes.c_uint32),
        ("size", ctypes.CFUNCTYPE(ctypes.c_int64, ctypes.c_void_p)),
        ("seek", ctypes.CFUNCTYPE(ctypes.c_int64, ctypes.c_void_p, ctypes.c_int64, SDL_IOWhence)),
        ("read", ctypes.CFUNCTYPE(ctypes.c_size_t, ctypes.c_void_p, ctypes.c_void_p, ctypes.c_size_t, ctypes.POINTER(SDL_IOStatus))),
        ("write", ctypes.CFUNCTYPE(ctypes.c_size_t, ctypes.c_void_p, ctypes.c_void_p, ctypes.c_size_t, ctypes.POINTER(SDL_IOStatus))),
        ("flush", ctypes.CFUNCTYPE(ctypes.c_bool, ctypes.c_void_p, ctypes.POINTER(SDL_IOStatus))),
        ("close", ctypes.CFUNCTYPE(ctypes.c_bool, ctypes.c_void_p))
    ]

class SDL_IOStream(ctypes.c_void_p):
    ...

SDL_FUNC("SDL_IOFromFile", ctypes.POINTER(SDL_IOStream), ctypes.c_char_p, ctypes.c_char_p)

SDL_PROP_IOSTREAM_WINDOWS_HANDLE_POINTER = "SDL.iostream.windows.handle"
SDL_PROP_IOSTREAM_STDIO_FILE_POINTER = "SDL.iostream.stdio.file"
SDL_PROP_IOSTREAM_FILE_DESCRIPTOR_NUMBER = "SDL.iostream.file_descriptor"
SDL_PROP_IOSTREAM_ANDROID_AASSET_POINTER = "SDL.iostream.android.aasset"

SDL_FUNC("SDL_IOFromMem", ctypes.POINTER(SDL_IOStream), ctypes.c_void_p, ctypes.c_size_t)

SDL_PROP_IOSTREAM_MEMORY_POINTER = "SDL.iostream.memory.base"
SDL_PROP_IOSTREAM_MEMORY_SIZE_NUMBER = "SDL.iostream.memory.size"

SDL_FUNC("SDL_IOFromConstMem", ctypes.POINTER(SDL_IOStream), ctypes.c_void_p, ctypes.c_size_t)
SDL_FUNC("SDL_IOFromDynamicMem", ctypes.POINTER(SDL_IOStream))

SDL_PROP_IOSTREAM_DYNAMIC_MEMORY_POINTER = "SDL.iostream.dynamic.memory"
SDL_PROP_IOSTREAM_DYNAMIC_CHUNKSIZE_NUMBER = "SDL.iostream.dynamic.chunksize"

SDL_FUNC("SDL_OpenIO", ctypes.POINTER(SDL_IOStream), ctypes.POINTER(SDL_IOStreamInterface), ctypes.c_void_p)
SDL_FUNC("SDL_CloseIO", ctypes.c_bool, ctypes.POINTER(SDL_IOStream))

SDL_FUNC("SDL_GetIOProperties", SDL_PropertiesID, ctypes.POINTER(SDL_IOStream))
SDL_FUNC("SDL_GetIOStatus", SDL_IOStatus, ctypes.POINTER(SDL_IOStream))
SDL_FUNC("SDL_GetIOSize", ctypes.c_int64, ctypes.POINTER(SDL_IOStream))

SDL_FUNC("SDL_SeekIO", ctypes.c_int64, ctypes.POINTER(SDL_IOStream), ctypes.c_int64, SDL_IOWhence)
SDL_FUNC("SDL_TellIO", ctypes.c_int64, ctypes.POINTER(SDL_IOStream))
SDL_FUNC("SDL_ReadIO", ctypes.c_size_t, ctypes.POINTER(SDL_IOStream), ctypes.c_void_p, ctypes.c_size_t)
SDL_FUNC("SDL_WriteIO", ctypes.c_size_t, ctypes.POINTER(SDL_IOStream), ctypes.c_void_p, ctypes.c_size_t)

SDL_FUNC("SDL_IOprintf", ctypes.c_size_t, ctypes.POINTER(SDL_IOStream), ctypes.c_char_p)
SDL_FUNC("SDL_IOvprintf", ctypes.c_size_t, ctypes.POINTER(SDL_IOStream), ctypes.c_char_p, ctypes.c_void_p)

SDL_FUNC("SDL_FlushIO", ctypes.c_bool, ctypes.POINTER(SDL_IOStream))

SDL_FUNC("SDL_LoadFile_IO", ctypes.c_void_p, ctypes.POINTER(SDL_IOStream), ctypes.POINTER(ctypes.c_size_t), ctypes.c_bool)
SDL_FUNC("SDL_LoadFile", ctypes.c_void_p, ctypes.c_char_p, ctypes.POINTER(ctypes.c_size_t))

SDL_FUNC("SDL_SaveFile_IO", ctypes.c_bool, ctypes.POINTER(SDL_IOStream), ctypes.POINTER(ctypes.c_void_p), ctypes.c_size_t, ctypes.c_bool)
SDL_FUNC("SDL_SaveFile", ctypes.c_bool, ctypes.c_char_p, ctypes.POINTER(ctypes.c_void_p), ctypes.c_size_t)

SDL_FUNC("SDL_ReadU8", ctypes.c_bool, ctypes.POINTER(SDL_IOStream), ctypes.POINTER(ctypes.c_uint8))
SDL_FUNC("SDL_ReadS8", ctypes.c_bool, ctypes.POINTER(SDL_IOStream), ctypes.POINTER(ctypes.c_int8))
SDL_FUNC("SDL_ReadU16LE", ctypes.c_bool, ctypes.POINTER(SDL_IOStream), ctypes.POINTER(ctypes.c_uint16))
SDL_FUNC("SDL_ReadS16LE", ctypes.c_bool, ctypes.POINTER(SDL_IOStream), ctypes.POINTER(ctypes.c_int16))
SDL_FUNC("SDL_ReadU16BE", ctypes.c_bool, ctypes.POINTER(SDL_IOStream), ctypes.POINTER(ctypes.c_uint16))
SDL_FUNC("SDL_ReadS16BE", ctypes.c_bool, ctypes.POINTER(SDL_IOStream), ctypes.POINTER(ctypes.c_int16))
SDL_FUNC("SDL_ReadU32LE", ctypes.c_bool, ctypes.POINTER(SDL_IOStream), ctypes.POINTER(ctypes.c_uint32))
SDL_FUNC("SDL_ReadS32LE", ctypes.c_bool, ctypes.POINTER(SDL_IOStream), ctypes.POINTER(ctypes.c_int32))
SDL_FUNC("SDL_ReadU32BE", ctypes.c_bool, ctypes.POINTER(SDL_IOStream), ctypes.POINTER(ctypes.c_uint32))
SDL_FUNC("SDL_ReadS32BE", ctypes.c_bool, ctypes.POINTER(SDL_IOStream), ctypes.POINTER(ctypes.c_int32))
SDL_FUNC("SDL_ReadU64LE", ctypes.c_bool, ctypes.POINTER(SDL_IOStream), ctypes.POINTER(ctypes.c_uint64))
SDL_FUNC("SDL_ReadS64LE", ctypes.c_bool, ctypes.POINTER(SDL_IOStream), ctypes.POINTER(ctypes.c_int64))
SDL_FUNC("SDL_ReadU64BE", ctypes.c_bool, ctypes.POINTER(SDL_IOStream), ctypes.POINTER(ctypes.c_uint64))
SDL_FUNC("SDL_ReadS64BE", ctypes.c_bool, ctypes.POINTER(SDL_IOStream), ctypes.POINTER(ctypes.c_int64))

SDL_FUNC("SDL_WriteU8", ctypes.c_bool, ctypes.POINTER(SDL_IOStream), ctypes.c_uint8)
SDL_FUNC("SDL_WriteS8", ctypes.c_bool, ctypes.POINTER(SDL_IOStream), ctypes.c_int8)
SDL_FUNC("SDL_WriteU16LE", ctypes.c_bool, ctypes.POINTER(SDL_IOStream), ctypes.c_uint16)
SDL_FUNC("SDL_WriteS16LE", ctypes.c_bool, ctypes.POINTER(SDL_IOStream), ctypes.c_int16)
SDL_FUNC("SDL_WriteU16BE", ctypes.c_bool, ctypes.POINTER(SDL_IOStream), ctypes.c_uint16)
SDL_FUNC("SDL_WriteS16BE", ctypes.c_bool, ctypes.POINTER(SDL_IOStream), ctypes.c_int16)
SDL_FUNC("SDL_WriteU32LE", ctypes.c_bool, ctypes.POINTER(SDL_IOStream), ctypes.c_uint32)
SDL_FUNC("SDL_WriteS32LE", ctypes.c_bool, ctypes.POINTER(SDL_IOStream), ctypes.c_int32)
SDL_FUNC("SDL_WriteU32BE", ctypes.c_bool, ctypes.POINTER(SDL_IOStream), ctypes.c_uint32)
SDL_FUNC("SDL_WriteS32BE", ctypes.c_bool, ctypes.POINTER(SDL_IOStream), ctypes.c_int32)
SDL_FUNC("SDL_WriteU64LE", ctypes.c_bool, ctypes.POINTER(SDL_IOStream), ctypes.c_uint64)
SDL_FUNC("SDL_WriteS64LE", ctypes.c_bool, ctypes.POINTER(SDL_IOStream), ctypes.c_int64)
SDL_FUNC("SDL_WriteU64BE", ctypes.c_bool, ctypes.POINTER(SDL_IOStream), ctypes.c_uint64)
SDL_FUNC("SDL_WriteS64BE", ctypes.c_bool, ctypes.POINTER(SDL_IOStream), ctypes.c_int64)


import ctypes, imgui, colorsys, time, sdl3, OpenGL.GL as gl

from imgui.integrations.opengl import ProgrammablePipelineRenderer

class SDL3Renderer(ProgrammablePipelineRenderer):
    """Basic SDL3 integration implementation."""
    MOUSE_WHEEL_OFFSET_SCALE = 0.5

    @staticmethod
    def setCustomStyle() -> None:
        style = imgui.get_style()
        style.colors[imgui.COLOR_TEXT] = imgui.Vec4(1.00, 1.00, 1.00, 1.00)
        style.colors[imgui.COLOR_TEXT_DISABLED] = imgui.Vec4(0.50, 0.50, 0.50, 1.00)
        style.colors[imgui.COLOR_WINDOW_BACKGROUND] = imgui.Vec4(0.00, 0.00, 0.00, 0.39)
        style.colors[imgui.COLOR_CHILD_BACKGROUND] = imgui.Vec4(0.00, 0.00, 0.00, 0.00)
        style.colors[imgui.COLOR_POPUP_BACKGROUND] = imgui.Vec4(0.00, 0.00, 0.00, 0.63)
        style.colors[imgui.COLOR_BORDER] = imgui.Vec4(1.00, 1.00, 1.00, 0.31)
        style.colors[imgui.COLOR_BORDER_SHADOW] = imgui.Vec4(0.00, 0.00, 0.00, 0.00)
        style.colors[imgui.COLOR_FRAME_BACKGROUND] = imgui.Vec4(0.00, 0.00, 0.00, 0.63)
        style.colors[imgui.COLOR_FRAME_BACKGROUND_HOVERED] = imgui.Vec4(0.23, 0.23, 0.23, 0.63)
        style.colors[imgui.COLOR_FRAME_BACKGROUND_ACTIVE] = imgui.Vec4(0.19, 0.19, 0.19, 0.39)
        style.colors[imgui.COLOR_TITLE_BACKGROUND] = imgui.Vec4(0.00, 0.00, 0.00, 0.63)
        style.colors[imgui.COLOR_TITLE_BACKGROUND_ACTIVE] = imgui.Vec4(0.00, 0.00, 0.00, 1.00)
        style.colors[imgui.COLOR_TITLE_BACKGROUND_COLLAPSED] = imgui.Vec4(0.00, 0.00, 0.00, 0.35)
        style.colors[imgui.COLOR_MENUBAR_BACKGROUND] = imgui.Vec4(0.00, 0.00, 0.00, 0.78)
        style.colors[imgui.COLOR_SCROLLBAR_BACKGROUND] = imgui.Vec4(0.05, 0.05, 0.05, 0.54)
        style.colors[imgui.COLOR_SCROLLBAR_GRAB] = imgui.Vec4(0.34, 0.34, 0.34, 0.63)
        style.colors[imgui.COLOR_SCROLLBAR_GRAB_HOVERED] = imgui.Vec4(0.50, 0.50, 0.50, 0.63)
        style.colors[imgui.COLOR_SCROLLBAR_GRAB_ACTIVE] = imgui.Vec4(0.38, 0.38, 0.38, 0.63)
        style.colors[imgui.COLOR_CHECK_MARK] = imgui.Vec4(0.22, 0.55, 0.74, 1.00)
        style.colors[imgui.COLOR_SLIDER_GRAB] = imgui.Vec4(0.34, 0.34, 0.34, 0.54)
        style.colors[imgui.COLOR_SLIDER_GRAB_ACTIVE] = imgui.Vec4(0.56, 0.56, 0.56, 0.54)
        style.colors[imgui.COLOR_BUTTON] = imgui.Vec4(0.00, 0.00, 0.00, 0.63)
        style.colors[imgui.COLOR_BUTTON_HOVERED] = imgui.Vec4(0.20, 0.22, 0.23, 0.63)
        style.colors[imgui.COLOR_BUTTON_ACTIVE] = imgui.Vec4(0.19, 0.19, 0.19, 0.39)
        style.colors[imgui.COLOR_HEADER] = imgui.Vec4(0.20, 0.20, 0.20, 0.78)
        style.colors[imgui.COLOR_HEADER_HOVERED] = imgui.Vec4(0.29, 0.29, 0.29, 0.78)
        style.colors[imgui.COLOR_HEADER_ACTIVE] = imgui.Vec4(0.19, 0.19, 0.19, 0.15)
        style.colors[imgui.COLOR_SEPARATOR] = imgui.Vec4(0.28, 0.28, 0.28, 0.29)
        style.colors[imgui.COLOR_SEPARATOR_HOVERED] = imgui.Vec4(0.44, 0.44, 0.44, 0.29)
        style.colors[imgui.COLOR_SEPARATOR_ACTIVE] = imgui.Vec4(0.40, 0.44, 0.47, 1.00)
        style.colors[imgui.COLOR_RESIZE_GRIP] = imgui.Vec4(0.28, 0.28, 0.28, 0.29)
        style.colors[imgui.COLOR_RESIZE_GRIP_HOVERED] = imgui.Vec4(0.44, 0.44, 0.44, 0.29)
        style.colors[imgui.COLOR_RESIZE_GRIP_ACTIVE] = imgui.Vec4(0.40, 0.44, 0.47, 1.00)
        style.colors[imgui.COLOR_TAB] = imgui.Vec4(0.00, 0.00, 0.00, 0.52)
        style.colors[imgui.COLOR_TAB_HOVERED] = imgui.Vec4(0.20, 0.20, 0.20, 0.36)
        style.colors[imgui.COLOR_TAB_ACTIVE] = imgui.Vec4(0.14, 0.14, 0.14, 1.00)
        style.colors[imgui.COLOR_TAB_UNFOCUSED] = imgui.Vec4(0.00, 0.00, 0.00, 0.52)
        style.colors[imgui.COLOR_TAB_UNFOCUSED_ACTIVE] = imgui.Vec4(0.14, 0.14, 0.14, 1.00)
        style.colors[imgui.COLOR_PLOT_LINES] = imgui.Vec4(1.00, 0.00, 0.00, 1.00)
        style.colors[imgui.COLOR_PLOT_LINES_HOVERED] = imgui.Vec4(1.00, 0.00, 0.00, 1.00)
        style.colors[imgui.COLOR_PLOT_HISTOGRAM] = imgui.Vec4(1.00, 0.00, 0.00, 1.00)
        style.colors[imgui.COLOR_PLOT_HISTOGRAM_HOVERED] = imgui.Vec4(1.00, 0.00, 0.00, 1.00)
        style.colors[imgui.COLOR_TABLE_HEADER_BACKGROUND] = imgui.Vec4(0.00, 0.00, 0.00, 0.52)
        style.colors[imgui.COLOR_TABLE_BORDER_STRONG] = imgui.Vec4(0.00, 0.00, 0.00, 0.52)
        style.colors[imgui.COLOR_TABLE_BORDER_LIGHT] = imgui.Vec4(0.28, 0.28, 0.28, 0.29)
        style.colors[imgui.COLOR_TABLE_ROW_BACKGROUND] = imgui.Vec4(0.00, 0.00, 0.00, 0.00)
        style.colors[imgui.COLOR_TABLE_ROW_BACKGROUND_ALT] = imgui.Vec4(1.00, 1.00, 1.00, 0.06)
        style.colors[imgui.COLOR_TEXT_SELECTED_BACKGROUND] = imgui.Vec4(0.20, 0.22, 0.23, 1.00)
        style.colors[imgui.COLOR_DRAG_DROP_TARGET] = imgui.Vec4(0.33, 0.67, 0.86, 1.00)
        style.colors[imgui.COLOR_NAV_HIGHLIGHT] = imgui.Vec4(1.00, 0.00, 0.00, 1.00)
        style.colors[imgui.COLOR_NAV_WINDOWING_HIGHLIGHT] = imgui.Vec4(1.00, 0.00, 0.00, 0.70)
        style.colors[imgui.COLOR_NAV_WINDOWING_DIM_BACKGROUND] = imgui.Vec4(1.00, 0.00, 0.00, 0.20)
        style.colors[imgui.COLOR_MODAL_WINDOW_DIM_BACKGROUND] = imgui.Vec4(1.00, 0.00, 0.00, 0.35)

        style.window_padding = imgui.Vec2(6.0, 6.0)
        style.frame_padding = imgui.Vec2(5.0, 2.0)
        style.cell_padding = imgui.Vec2(6.0, 6.0)
        style.item_spacing = imgui.Vec2(6.0, 6.0)
        style.item_inner_spacing = imgui.Vec2(6.0, 6.0)
        style.touch_extra_padding = imgui.Vec2(0.0, 0.0)
        style.window_title_align = imgui.Vec2(0.5, 0.5)
        style.indent_spacing = 25.0
        style.scrollbar_size = 10.0
        style.grab_min_size = 10.0
        style.window_border_size = 1.0
        style.child_border_size = 1.0
        style.popup_border_size = 1.0
        style.frame_border_size = 1.0
        style.tab_border_size = 1.0
        style.window_rounding = 5.0
        style.child_rounding = 4.0
        style.frame_rounding = 2.0
        style.popup_rounding = 4.0
        style.scrollbar_rounding = 9.0
        style.grab_rounding = 2.0
        style.log_slider_deadzone = 4.0
        style.tab_rounding = 4.0

    def __init__(self, window: sdl3.LP_SDL_Window) -> None:
        super(SDL3Renderer, self).__init__()
        
        self.window = window
        self.lastTime = sdl3.SDL_GetTicks() / 1000.0
        self.io.get_clipboard_text_fn = lambda: sdl3.SDL_GetClipboardText()
        self.io.set_clipboard_text_fn = lambda text: sdl3.SDL_SetClipboardText(text.encode())

        self.io.key_map[imgui.KEY_TAB] = sdl3.SDL_SCANCODE_TAB
        self.io.key_map[imgui.KEY_LEFT_ARROW] = sdl3.SDL_SCANCODE_LEFT
        self.io.key_map[imgui.KEY_RIGHT_ARROW] = sdl3.SDL_SCANCODE_RIGHT
        self.io.key_map[imgui.KEY_UP_ARROW] = sdl3.SDL_SCANCODE_UP
        self.io.key_map[imgui.KEY_DOWN_ARROW] = sdl3.SDL_SCANCODE_DOWN
        self.io.key_map[imgui.KEY_PAGE_UP] = sdl3.SDL_SCANCODE_PAGEUP
        self.io.key_map[imgui.KEY_PAGE_DOWN] = sdl3.SDL_SCANCODE_PAGEDOWN
        self.io.key_map[imgui.KEY_HOME] = sdl3.SDL_SCANCODE_HOME
        self.io.key_map[imgui.KEY_END] = sdl3.SDL_SCANCODE_END
        self.io.key_map[imgui.KEY_INSERT] = sdl3.SDL_SCANCODE_INSERT
        self.io.key_map[imgui.KEY_DELETE] = sdl3.SDL_SCANCODE_DELETE
        self.io.key_map[imgui.KEY_BACKSPACE] = sdl3.SDL_SCANCODE_BACKSPACE
        self.io.key_map[imgui.KEY_SPACE] = sdl3.SDL_SCANCODE_SPACE
        self.io.key_map[imgui.KEY_ENTER] = sdl3.SDL_SCANCODE_RETURN
        self.io.key_map[imgui.KEY_ESCAPE] = sdl3.SDL_SCANCODE_ESCAPE
        self.io.key_map[imgui.KEY_PAD_ENTER] = sdl3.SDL_SCANCODE_KP_ENTER
        self.io.key_map[imgui.KEY_A] = sdl3.SDL_SCANCODE_A
        self.io.key_map[imgui.KEY_C] = sdl3.SDL_SCANCODE_C
        self.io.key_map[imgui.KEY_V] = sdl3.SDL_SCANCODE_V
        self.io.key_map[imgui.KEY_X] = sdl3.SDL_SCANCODE_X
        self.io.key_map[imgui.KEY_Y] = sdl3.SDL_SCANCODE_Y
        self.io.key_map[imgui.KEY_Z] = sdl3.SDL_SCANCODE_Z

    def processEvent(self, event: sdl3.SDL_Event) -> None:
        if event.type in [sdl3.SDL_EVENT_MOUSE_WHEEL]:
            self.io.mouse_wheel = event.wheel.y * self.MOUSE_WHEEL_OFFSET_SCALE

        if event.type in [sdl3.SDL_EVENT_MOUSE_BUTTON_UP, sdl3.SDL_EVENT_MOUSE_BUTTON_DOWN]:
            buttons = [sdl3.SDL_BUTTON_LEFT, sdl3.SDL_BUTTON_RIGHT, sdl3.SDL_BUTTON_MIDDLE]

            if event.button.button in buttons:
                self.io.mouse_down[buttons.index(event.button.button)] = event.type == sdl3.SDL_EVENT_MOUSE_BUTTON_DOWN

        if event.type in [sdl3.SDL_EVENT_MOUSE_MOTION]:
            self.io.mouse_pos = (event.motion.x, event.motion.y) \
                if sdl3.SDL_GetWindowFlags(self.window) & sdl3.SDL_WINDOW_MOUSE_FOCUS else (-1, -1)

        if event.type in [sdl3.SDL_EVENT_KEY_UP, sdl3.SDL_EVENT_KEY_DOWN]:
            if event.key.scancode < sdl3.SDL_SCANCODE_COUNT:
                self.io.keys_down[event.key.scancode] = event.type == sdl3.SDL_EVENT_KEY_DOWN

            self.io.key_shift = (sdl3.SDL_GetModState() & sdl3.SDL_KMOD_SHIFT) != 0
            self.io.key_ctrl = (sdl3.SDL_GetModState() & sdl3.SDL_KMOD_CTRL) != 0
            self.io.key_alt = (sdl3.SDL_GetModState() & sdl3.SDL_KMOD_ALT) != 0
            self.io.key_super = (sdl3.SDL_GetModState() & sdl3.SDL_KMOD_GUI) != 0

        if event.type == sdl3.SDL_EVENT_TEXT_INPUT:
            for char in event.text.text.decode("utf-8"):
                self.io.add_input_character(ord(char))

    def processInputs(self) -> None:
        """This function should be called before processing events."""

        width, height = ctypes.c_int(0), ctypes.c_int(0)
        sdl3.SDL_GetWindowSize(self.window, ctypes.byref(width), ctypes.byref(height))
        self.io.display_size, self.io.display_fb_scale = (width.value, height.value), (1, 1)
        self.io.mouse_wheel = 0
        
        currentTime = sdl3.SDL_GetTicks() / 1000.0; deltaTime = currentTime - self.lastTime
        self.io.delta_time = 1.0 / 1000.0 if deltaTime <= 0.0 else deltaTime
        self.lastTime = currentTime

@sdl3.SDL_main_func
def main(argc: ctypes.c_int, argv: sdl3.LP_c_char_p) -> ctypes.c_int:
    print(f"loaded {sum(len(v) for k, v in sdl3.modules.items())} functions.")

    if not sdl3.SDL_Init(sdl3.SDL_INIT_VIDEO | sdl3.SDL_INIT_EVENTS):
        print(f"failed to initialize library: {sdl3.SDL_GetError().decode().lower()}.")
        return -1
    
    sdl3.SDL_GL_SetAttribute(sdl3.SDL_GL_CONTEXT_MAJOR_VERSION, 4)
    sdl3.SDL_GL_SetAttribute(sdl3.SDL_GL_CONTEXT_MINOR_VERSION, 6)
    sdl3.SDL_GL_SetAttribute(sdl3.SDL_GL_CONTEXT_PROFILE_MASK, sdl3.SDL_GL_CONTEXT_PROFILE_COMPATIBILITY)
    sdl3.SDL_GL_SetAttribute(sdl3.SDL_GL_CONTEXT_FLAGS, sdl3.SDL_GL_CONTEXT_FORWARD_COMPATIBLE_FLAG)
    window = sdl3.SDL_CreateWindow("Aermoss".encode(), 1200, 600, sdl3.SDL_WINDOW_OPENGL | sdl3.SDL_WINDOW_RESIZABLE)

    if not window:
        print(f"failed to create window: {sdl3.SDL_GetError().decode().lower()}.", flush = True)
        return -1

    context = sdl3.SDL_GL_CreateContext(window)
    sdl3.SDL_GL_MakeCurrent(window, context)

    if not context:
        print(f"failed to create context: {sdl3.SDL_GetError().decode().lower()}.", flush = True)
        return -1

    imgui.create_context()
    imgui.get_io().ini_file_name = None
    SDL3Renderer.setCustomStyle()

    renderer = SDL3Renderer(window)
    running, hue, lastTime = True, 0.0, time.time()
    event = sdl3.SDL_Event()

    while running:
        renderer.processInputs()

        while sdl3.SDL_PollEvent(ctypes.byref(event)):
            renderer.processEvent(event)

            match event.type:
                case sdl3.SDL_EVENT_QUIT:
                    running = False

                case sdl3.SDL_EVENT_KEY_DOWN:
                    if event.key.key == sdl3.SDLK_ESCAPE:
                        running = False

        lastTime, deltaTime = \
            time.time(), time.time() - lastTime

        hue = (hue + 0.5 * deltaTime) % 360.0
        gl.glClearColor(*colorsys.hsv_to_rgb(hue, 1.0, 0.1), 1.0)
        gl.glClear(gl.GL_COLOR_BUFFER_BIT)
        imgui.new_frame()

        imgui.begin("Hello, World!")
        imgui.text("This is some useful text.")
        imgui.end()

        imgui.end_frame()
        imgui.render()
        renderer.render(imgui.get_draw_data())
        sdl3.SDL_GL_SwapWindow(window)

    renderer.shutdown()
    # imgui.destroy_context()

    sdl3.SDL_GL_MakeCurrent(window, None)
    sdl3.SDL_GL_DestroyContext(context)
    sdl3.SDL_DestroyWindow(window)
    sdl3.SDL_Quit()
    return 0


from .__init__ import ctypes, \
    SDL_FUNC, SDL_SET_CURRENT_BINARY, SDL_BINARY

SDL_SET_CURRENT_BINARY(SDL_BINARY)

class SDL_AsyncIO(ctypes.c_void_p):
    ...

SDL_AsyncIOTaskType = ctypes.c_int

SDL_ASYNCIO_TASK_READ = 0
SDL_ASYNCIO_TASK_WRITE = 1
SDL_ASYNCIO_TASK_CLOSE = 2

SDL_AsyncIOResult = ctypes.c_int

SDL_ASYNCIO_COMPLETE = 0
SDL_ASYNCIO_FAILURE = 1
SDL_ASYNCIO_CANCELED = 2

class SDL_AsyncIOOutcome(ctypes.Structure):
    _fields_ = [
        ("asyncio", ctypes.POINTER(SDL_AsyncIO)),
        ("type", SDL_AsyncIOTaskType),
        ("result", SDL_AsyncIOResult),
        ("buffer", ctypes.c_void_p),
        ("offset", ctypes.c_uint64),
        ("bytes_requested", ctypes.c_uint64),
        ("bytes_transferred", ctypes.c_uint64),
        ("userdata", ctypes.c_void_p)
    ]

class SDL_AsyncIOQueue(ctypes.c_void_p):
    ...

SDL_FUNC("SDL_AsyncIOFromFile", ctypes.POINTER(SDL_AsyncIO), ctypes.c_char_p, ctypes.c_char_p)
SDL_FUNC("SDL_GetAsyncIOSize", ctypes.c_int64, ctypes.POINTER(SDL_AsyncIO))

SDL_FUNC("SDL_ReadAsyncIO", ctypes.c_bool, ctypes.POINTER(SDL_AsyncIO), ctypes.c_void_p, ctypes.c_uint64, ctypes.c_uint64, ctypes.POINTER(SDL_AsyncIOQueue), ctypes.c_void_p)
SDL_FUNC("SDL_WriteAsyncIO", ctypes.c_bool, ctypes.POINTER(SDL_AsyncIO), ctypes.c_void_p, ctypes.c_uint64, ctypes.c_uint64, ctypes.POINTER(SDL_AsyncIOQueue), ctypes.c_void_p)
SDL_FUNC("SDL_CloseAsyncIO", ctypes.c_bool, ctypes.POINTER(SDL_AsyncIO), ctypes.c_bool, ctypes.POINTER(SDL_AsyncIOQueue), ctypes.c_void_p)

SDL_FUNC("SDL_CreateAsyncIOQueue", ctypes.POINTER(SDL_AsyncIOQueue))
SDL_FUNC("SDL_DestroyAsyncIOQueue", None, ctypes.POINTER(SDL_AsyncIOQueue))
SDL_FUNC("SDL_GetAsyncIOResult", ctypes.c_bool, ctypes.POINTER(SDL_AsyncIOQueue), ctypes.POINTER(SDL_AsyncIOOutcome))
SDL_FUNC("SDL_WaitAsyncIOResult", ctypes.c_bool, ctypes.POINTER(SDL_AsyncIOQueue), ctypes.POINTER(SDL_AsyncIOOutcome), ctypes.c_int32)
SDL_FUNC("SDL_SignalAsyncIOQueue", None, ctypes.POINTER(SDL_AsyncIOQueue))

SDL_FUNC("SDL_LoadFileAsync", ctypes.c_bool, ctypes.c_char_p, ctypes.POINTER(SDL_AsyncIOQueue), ctypes.c_void_p)


from .SDL_assert import *
from .SDL_asyncio import *
from .SDL_atomic import *
from .SDL_audio import *
from .SDL_bits import *
from .SDL_blendmode import *
from .SDL_camera import *
from .SDL_clipboard import *
from .SDL_cpuinfo import *
from .SDL_dialog import *
from .SDL_endian import *
from .SDL_error import *
from .SDL_events import *
from .SDL_filesystem import *
from .SDL_gamepad import *
from .SDL_gpu import *
from .SDL_guid import *
from .SDL_haptic import *
from .SDL_hidapi import *
from .SDL_hints import *
from .SDL_image import *
from .SDL_init import *
from .SDL_iostream import *
from .SDL_joystick import *
from .SDL_keyboard import *
from .SDL_keycode import *
from .SDL_loadso import *
from .SDL_locale import *
from .SDL_log import *
from .SDL_main_impl import *
from .SDL_main import *
from .SDL_messagebox import *
from .SDL_metal import *
from .SDL_misc import *
from .SDL_mixer import *
from .SDL_mouse import *
from .SDL_mutex import *
from .SDL_net import *
from .SDL_pen import *
from .SDL_pixels import *
from .SDL_platform import *
from .SDL_power import *
from .SDL_process import *
from .SDL_properties import *
from .SDL_rect import *
from .SDL_render import *
from .SDL_rtf import *
from .SDL_scancode import *
from .SDL_sensor import *
from .SDL_stdinc import *
from .SDL_storage import *
from .SDL_surface import *
from .SDL_system import *
from .SDL_textengine import *
from .SDL_thread import *
from .SDL_time import *
from .SDL_timer import *
from .SDL_touch import *
from .SDL_tray import *
from .SDL_ttf import *
from .SDL_version import *
from .SDL_video import *
from .SDL_vulkan import *


from .__init__ import ctypes, \
    SDL_FUNC, SDL_SET_CURRENT_BINARY, SDL_BINARY

SDL_SET_CURRENT_BINARY(SDL_BINARY)

SDL_Keycode = ctypes.c_uint32

SDLK_EXTENDED_MASK = 1 << 29
SDLK_SCANCODE_MASK = 1 << 30
SDL_SCANCODE_TO_KEYCODE = lambda x: x | SDLK_SCANCODE_MASK
SDLK_UNKNOWN = 0x00000000
SDLK_RETURN = 0x0000000d
SDLK_ESCAPE = 0x0000001b
SDLK_BACKSPACE = 0x00000008
SDLK_TAB = 0x00000009
SDLK_SPACE = 0x00000020
SDLK_EXCLAIM = 0x00000021
SDLK_DBLAPOSTROPHE = 0x00000022
SDLK_HASH = 0x00000023
SDLK_DOLLAR = 0x00000024
SDLK_PERCENT = 0x00000025
SDLK_AMPERSAND = 0x00000026
SDLK_APOSTROPHE = 0x00000027
SDLK_LEFTPAREN = 0x00000028
SDLK_RIGHTPAREN = 0x00000029
SDLK_ASTERISK = 0x0000002a
SDLK_PLUS = 0x0000002b
SDLK_COMMA = 0x0000002c
SDLK_MINUS = 0x0000002d
SDLK_PERIOD = 0x0000002e
SDLK_SLASH = 0x0000002f
SDLK_0 = 0x00000030
SDLK_1 = 0x00000031
SDLK_2 = 0x00000032
SDLK_3 = 0x00000033
SDLK_4 = 0x00000034
SDLK_5 = 0x00000035
SDLK_6 = 0x00000036
SDLK_7 = 0x00000037
SDLK_8 = 0x00000038
SDLK_9 = 0x00000039
SDLK_COLON = 0x0000003a
SDLK_SEMICOLON = 0x0000003b
SDLK_LESS = 0x0000003c
SDLK_EQUALS = 0x0000003d
SDLK_GREATER = 0x0000003e
SDLK_QUESTION = 0x0000003f
SDLK_AT = 0x00000040
SDLK_LEFTBRACKET = 0x0000005b
SDLK_BACKSLASH = 0x0000005c
SDLK_RIGHTBRACKET = 0x0000005d
SDLK_CARET = 0x0000005e
SDLK_UNDERSCORE = 0x0000005f
SDLK_GRAVE = 0x00000060
SDLK_A = 0x00000061
SDLK_B = 0x00000062
SDLK_C = 0x00000063
SDLK_D = 0x00000064
SDLK_E = 0x00000065
SDLK_F = 0x00000066
SDLK_G = 0x00000067
SDLK_H = 0x00000068
SDLK_I = 0x00000069
SDLK_J = 0x0000006a
SDLK_K = 0x0000006b
SDLK_L = 0x0000006c
SDLK_M = 0x0000006d
SDLK_N = 0x0000006e
SDLK_O = 0x0000006f
SDLK_P = 0x00000070
SDLK_Q = 0x00000071
SDLK_R = 0x00000072
SDLK_S = 0x00000073
SDLK_T = 0x00000074
SDLK_U = 0x00000075
SDLK_V = 0x00000076
SDLK_W = 0x00000077
SDLK_X = 0x00000078
SDLK_Y = 0x00000079
SDLK_Z = 0x0000007a
SDLK_LEFTBRACE = 0x0000007b
SDLK_PIPE = 0x0000007c
SDLK_RIGHTBRACE = 0x0000007d
SDLK_TILDE = 0x0000007e
SDLK_DELETE = 0x0000007f
SDLK_PLUSMINUS = 0x000000b1
SDLK_CAPSLOCK = 0x40000039
SDLK_F1 = 0x4000003a
SDLK_F2 = 0x4000003b
SDLK_F3 = 0x4000003c
SDLK_F4 = 0x4000003d
SDLK_F5 = 0x4000003e
SDLK_F6 = 0x4000003f
SDLK_F7 = 0x40000040
SDLK_F8 = 0x40000041
SDLK_F9 = 0x40000042
SDLK_F10 = 0x40000043
SDLK_F11 = 0x40000044
SDLK_F12 = 0x40000045
SDLK_PRINTSCREEN = 0x40000046
SDLK_SCROLLLOCK = 0x40000047
SDLK_PAUSE = 0x40000048
SDLK_INSERT = 0x40000049
SDLK_HOME = 0x4000004a
SDLK_PAGEUP = 0x4000004b
SDLK_END = 0x4000004d
SDLK_PAGEDOWN = 0x4000004e
SDLK_RIGHT = 0x4000004f
SDLK_LEFT = 0x40000050
SDLK_DOWN = 0x40000051
SDLK_UP = 0x40000052
SDLK_NUMLOCKCLEAR = 0x40000053
SDLK_KP_DIVIDE = 0x40000054
SDLK_KP_MULTIPLY = 0x40000055
SDLK_KP_MINUS = 0x40000056
SDLK_KP_PLUS = 0x40000057
SDLK_KP_ENTER = 0x40000058
SDLK_KP_1 = 0x40000059
SDLK_KP_2 = 0x4000005a
SDLK_KP_3 = 0x4000005b
SDLK_KP_4 = 0x4000005c
SDLK_KP_5 = 0x4000005d
SDLK_KP_6 = 0x4000005e
SDLK_KP_7 = 0x4000005f
SDLK_KP_8 = 0x40000060
SDLK_KP_9 = 0x40000061
SDLK_KP_0 = 0x40000062
SDLK_KP_PERIOD = 0x40000063
SDLK_APPLICATION = 0x40000065
SDLK_POWER = 0x40000066
SDLK_KP_EQUALS = 0x40000067
SDLK_F13 = 0x40000068
SDLK_F14 = 0x40000069
SDLK_F15 = 0x4000006a
SDLK_F16 = 0x4000006b
SDLK_F17 = 0x4000006c
SDLK_F18 = 0x4000006d
SDLK_F19 = 0x4000006e
SDLK_F20 = 0x4000006f
SDLK_F21 = 0x40000070
SDLK_F22 = 0x40000071
SDLK_F23 = 0x40000072
SDLK_F24 = 0x40000073
SDLK_EXECUTE = 0x40000074
SDLK_HELP = 0x40000075
SDLK_MENU = 0x40000076
SDLK_SELECT = 0x40000077
SDLK_STOP = 0x40000078
SDLK_AGAIN = 0x40000079
SDLK_UNDO = 0x4000007a
SDLK_CUT = 0x4000007b
SDLK_COPY = 0x4000007c
SDLK_PASTE = 0x4000007d
SDLK_FIND = 0x4000007e
SDLK_MUTE = 0x4000007f
SDLK_VOLUMEUP = 0x40000080
SDLK_VOLUMEDOWN = 0x40000081
SDLK_KP_COMMA = 0x40000085
SDLK_KP_EQUALSAS400 = 0x40000086
SDLK_ALTERASE = 0x40000099
SDLK_SYSREQ = 0x4000009a
SDLK_CANCEL = 0x4000009b
SDLK_CLEAR = 0x4000009c
SDLK_PRIOR = 0x4000009d
SDLK_RETURN2 = 0x4000009e
SDLK_SEPARATOR = 0x4000009f
SDLK_OUT = 0x400000a0
SDLK_OPER = 0x400000a1
SDLK_CLEARAGAIN = 0x400000a2
SDLK_CRSEL = 0x400000a3
SDLK_EXSEL = 0x400000a4
SDLK_KP_00 = 0x400000b0
SDLK_KP_000 = 0x400000b1
SDLK_THOUSANDSSEPARATOR = 0x400000b2
SDLK_DECIMALSEPARATOR = 0x400000b3
SDLK_CURRENCYUNIT = 0x400000b4
SDLK_CURRENCYSUBUNIT = 0x400000b5
SDLK_KP_LEFTPAREN = 0x400000b6
SDLK_KP_RIGHTPAREN = 0x400000b7
SDLK_KP_LEFTBRACE = 0x400000b8
SDLK_KP_RIGHTBRACE = 0x400000b9
SDLK_KP_TAB = 0x400000ba
SDLK_KP_BACKSPACE = 0x400000bb
SDLK_KP_A = 0x400000bc
SDLK_KP_B = 0x400000bd
SDLK_KP_C = 0x400000be
SDLK_KP_D = 0x400000bf
SDLK_KP_E = 0x400000c0
SDLK_KP_F = 0x400000c1
SDLK_KP_XOR = 0x400000c2
SDLK_KP_POWER = 0x400000c3
SDLK_KP_PERCENT = 0x400000c4
SDLK_KP_LESS = 0x400000c5
SDLK_KP_GREATER = 0x400000c6
SDLK_KP_AMPERSAND = 0x400000c7
SDLK_KP_DBLAMPERSAND = 0x400000c8
SDLK_KP_VERTICALBAR = 0x400000c9
SDLK_KP_DBLVERTICALBAR = 0x400000ca
SDLK_KP_COLON = 0x400000cb
SDLK_KP_HASH = 0x400000cc
SDLK_KP_SPACE = 0x400000cd
SDLK_KP_AT = 0x400000ce
SDLK_KP_EXCLAM = 0x400000cf
SDLK_KP_MEMSTORE = 0x400000d0
SDLK_KP_MEMRECALL = 0x400000d1
SDLK_KP_MEMCLEAR = 0x400000d2
SDLK_KP_MEMADD = 0x400000d3
SDLK_KP_MEMSUBTRACT = 0x400000d4
SDLK_KP_MEMMULTIPLY = 0x400000d5
SDLK_KP_MEMDIVIDE = 0x400000d6
SDLK_KP_PLUSMINUS = 0x400000d7
SDLK_KP_CLEAR = 0x400000d8
SDLK_KP_CLEARENTRY = 0x400000d9
SDLK_KP_BINARY = 0x400000da
SDLK_KP_OCTAL = 0x400000db
SDLK_KP_DECIMAL = 0x400000dc
SDLK_KP_HEXADECIMAL = 0x400000dd
SDLK_LCTRL = 0x400000e0
SDLK_LSHIFT = 0x400000e1
SDLK_LALT = 0x400000e2
SDLK_LGUI = 0x400000e3
SDLK_RCTRL = 0x400000e4
SDLK_RSHIFT = 0x400000e5
SDLK_RALT = 0x400000e6
SDLK_RGUI = 0x400000e7
SDLK_MODE = 0x40000101
SDLK_SLEEP = 0x40000102
SDLK_WAKE = 0x40000103
SDLK_CHANNEL_INCREMENT = 0x40000104
SDLK_CHANNEL_DECREMENT = 0x40000105
SDLK_MEDIA_PLAY = 0x40000106
SDLK_MEDIA_PAUSE = 0x40000107
SDLK_MEDIA_RECORD = 0x40000108
SDLK_MEDIA_FAST_FORWARD = 0x40000109
SDLK_MEDIA_REWIND = 0x4000010a
SDLK_MEDIA_NEXT_TRACK = 0x4000010b
SDLK_MEDIA_PREVIOUS_TRACK = 0x4000010c
SDLK_MEDIA_STOP = 0x4000010d
SDLK_MEDIA_EJECT = 0x4000010e
SDLK_MEDIA_PLAY_PAUSE = 0x4000010f
SDLK_MEDIA_SELECT = 0x40000110
SDLK_AC_NEW = 0x40000111
SDLK_AC_OPEN = 0x40000112
SDLK_AC_CLOSE = 0x40000113
SDLK_AC_EXIT = 0x40000114
SDLK_AC_SAVE = 0x40000115
SDLK_AC_PRINT = 0x40000116
SDLK_AC_PROPERTIES = 0x40000117
SDLK_AC_SEARCH = 0x40000118
SDLK_AC_HOME = 0x40000119
SDLK_AC_BACK = 0x4000011a
SDLK_AC_FORWARD = 0x4000011b
SDLK_AC_STOP = 0x4000011c
SDLK_AC_REFRESH = 0x4000011d
SDLK_AC_BOOKMARKS = 0x4000011e
SDLK_SOFTLEFT = 0x4000011f
SDLK_SOFTRIGHT = 0x40000120
SDLK_CALL = 0x40000121
SDLK_ENDCALL = 0x40000122
SDLK_LEFT_TAB = 0x20000001
SDLK_LEVEL5_SHIFT = 0x20000002
SDLK_MULTI_KEY_COMPOSE = 0x20000003
SDLK_LMETA = 0x20000004
SDLK_RMETA = 0x20000005
SDLK_LHYPER = 0x20000006
SDLK_RHYPER = 0x20000007

SDL_Keymod = ctypes.c_uint16

SDL_KMOD_NONE = 0x0000
SDL_KMOD_LSHIFT = 0x0001
SDL_KMOD_RSHIFT = 0x0002
SDL_KMOD_LEVEL5 = 0x0004
SDL_KMOD_LCTRL = 0x0040
SDL_KMOD_RCTRL = 0x0080
SDL_KMOD_LALT = 0x0100
SDL_KMOD_RALT = 0x0200
SDL_KMOD_LGUI = 0x0400
SDL_KMOD_RGUI = 0x0800
SDL_KMOD_NUM = 0x1000
SDL_KMOD_CAPS = 0x2000
SDL_KMOD_MODE = 0x4000
SDL_KMOD_SCROLL = 0x8000
SDL_KMOD_CTRL = SDL_KMOD_LCTRL | SDL_KMOD_RCTRL
SDL_KMOD_SHIFT = SDL_KMOD_LSHIFT | SDL_KMOD_RSHIFT
SDL_KMOD_ALT = SDL_KMOD_LALT | SDL_KMOD_RALT
SDL_KMOD_GUI = SDL_KMOD_LGUI | SDL_KMOD_RGUI


from .__init__ import ctypes, \
    SDL_FUNC, SDL_SET_CURRENT_BINARY, SDL_BINARY

from .SDL_events import SDL_Event

SDL_SET_CURRENT_BINARY(SDL_BINARY)

SDL_InitFlags = ctypes.c_uint32

SDL_INIT_AUDIO = 0x00000010
SDL_INIT_VIDEO = 0x00000020
SDL_INIT_JOYSTICK = 0x00000200
SDL_INIT_HAPTIC = 0x00001000
SDL_INIT_GAMEPAD = 0x00002000
SDL_INIT_EVENTS = 0x00004000
SDL_INIT_SENSOR = 0x00008000
SDL_INIT_CAMERA = 0x00010000

SDL_AppResult = ctypes.c_int

SDL_APP_CONTINUE = 0
SDL_APP_SUCCESS = 1
SDL_APP_FAILURE = 2

SDL_AppInit_func = ctypes.CFUNCTYPE(SDL_AppResult, ctypes.POINTER(ctypes.c_void_p), ctypes.c_int, ctypes.POINTER(ctypes.c_char_p))
SDL_AppIterate_func = ctypes.CFUNCTYPE(SDL_AppResult, ctypes.c_void_p)
SDL_AppEvent_func = ctypes.CFUNCTYPE(SDL_AppResult, ctypes.c_void_p, ctypes.POINTER(SDL_Event))
SDL_AppQuit_func = ctypes.CFUNCTYPE(None, ctypes.c_void_p, SDL_AppResult)

SDL_FUNC("SDL_Init", ctypes.c_bool, SDL_InitFlags)
SDL_FUNC("SDL_InitSubSystem", ctypes.c_bool, SDL_InitFlags)
SDL_FUNC("SDL_QuitSubSystem", None, SDL_InitFlags)
SDL_FUNC("SDL_WasInit", SDL_InitFlags, SDL_InitFlags)
SDL_FUNC("SDL_Quit", None)

SDL_FUNC("SDL_IsMainThread", ctypes.c_bool)
SDL_MainThreadCallback = ctypes.CFUNCTYPE(None, ctypes.c_void_p)
SDL_FUNC("SDL_RunOnMainThread", ctypes.c_bool, SDL_MainThreadCallback, ctypes.c_void_p, ctypes.c_bool)

SDL_FUNC("SDL_SetAppMetadata", ctypes.c_bool, ctypes.c_char_p, ctypes.c_char_p, ctypes.c_char_p)
SDL_FUNC("SDL_SetAppMetadataProperty", ctypes.c_bool, ctypes.c_char_p, ctypes.c_char_p)

SDL_PROP_APP_METADATA_NAME_STRING = "SDL.app.metadata.name"
SDL_PROP_APP_METADATA_VERSION_STRING = "SDL.app.metadata.version"
SDL_PROP_APP_METADATA_IDENTIFIER_STRING = "SDL.app.metadata.identifier"
SDL_PROP_APP_METADATA_CREATOR_STRING = "SDL.app.metadata.creator"
SDL_PROP_APP_METADATA_COPYRIGHT_STRING = "SDL.app.metadata.copyright"
SDL_PROP_APP_METADATA_URL_STRING = "SDL.app.metadata.url"
SDL_PROP_APP_METADATA_TYPE_STRING = "SDL.app.metadata.type"

SDL_FUNC("SDL_GetAppMetadataProperty", ctypes.c_char_p, ctypes.c_char_p)


from .__init__ import ctypes, \
    SDL_FUNC, SDL_SET_CURRENT_BINARY, SDL_BINARY

SDL_SET_CURRENT_BINARY(SDL_BINARY)

SDL_HINT_ALLOW_ALT_TAB_WHILE_GRABBED = "SDL_ALLOW_ALT_TAB_WHILE_GRABBED"
SDL_HINT_ANDROID_ALLOW_RECREATE_ACTIVITY = "SDL_ANDROID_ALLOW_RECREATE_ACTIVITY"
SDL_HINT_ANDROID_BLOCK_ON_PAUSE = "SDL_ANDROID_BLOCK_ON_PAUSE"
SDL_HINT_ANDROID_LOW_LATENCY_AUDIO = "SDL_ANDROID_LOW_LATENCY_AUDIO"
SDL_HINT_ANDROID_TRAP_BACK_BUTTON = "SDL_ANDROID_TRAP_BACK_BUTTON"
SDL_HINT_APP_ID = "SDL_APP_ID"
SDL_HINT_APP_NAME = "SDL_APP_NAME"
SDL_HINT_APPLE_TV_CONTROLLER_UI_EVENTS = "SDL_APPLE_TV_CONTROLLER_UI_EVENTS"
SDL_HINT_APPLE_TV_REMOTE_ALLOW_ROTATION = "SDL_APPLE_TV_REMOTE_ALLOW_ROTATION"
SDL_HINT_AUDIO_ALSA_DEFAULT_DEVICE = "SDL_AUDIO_ALSA_DEFAULT_DEVICE"
SDL_HINT_AUDIO_ALSA_DEFAULT_PLAYBACK_DEVICE = "SDL_AUDIO_ALSA_DEFAULT_PLAYBACK_DEVICE"
SDL_HINT_AUDIO_ALSA_DEFAULT_RECORDING_DEVICE = "SDL_AUDIO_ALSA_DEFAULT_RECORDING_DEVICE"
SDL_HINT_AUDIO_CATEGORY = "SDL_AUDIO_CATEGORY"
SDL_HINT_AUDIO_CHANNELS = "SDL_AUDIO_CHANNELS"
SDL_HINT_AUDIO_DEVICE_APP_ICON_NAME = "SDL_AUDIO_DEVICE_APP_ICON_NAME"
SDL_HINT_AUDIO_DEVICE_SAMPLE_FRAMES = "SDL_AUDIO_DEVICE_SAMPLE_FRAMES"
SDL_HINT_AUDIO_DEVICE_STREAM_NAME = "SDL_AUDIO_DEVICE_STREAM_NAME"
SDL_HINT_AUDIO_DEVICE_STREAM_ROLE = "SDL_AUDIO_DEVICE_STREAM_ROLE"
SDL_HINT_AUDIO_DISK_INPUT_FILE = "SDL_AUDIO_DISK_INPUT_FILE"
SDL_HINT_AUDIO_DISK_OUTPUT_FILE = "SDL_AUDIO_DISK_OUTPUT_FILE"
SDL_HINT_AUDIO_DISK_TIMESCALE = "SDL_AUDIO_DISK_TIMESCALE"
SDL_HINT_AUDIO_DRIVER = "SDL_AUDIO_DRIVER"
SDL_HINT_AUDIO_DUMMY_TIMESCALE = "SDL_AUDIO_DUMMY_TIMESCALE"
SDL_HINT_AUDIO_FORMAT = "SDL_AUDIO_FORMAT"
SDL_HINT_AUDIO_FREQUENCY = "SDL_AUDIO_FREQUENCY"
SDL_HINT_AUDIO_INCLUDE_MONITORS = "SDL_AUDIO_INCLUDE_MONITORS"
SDL_HINT_AUTO_UPDATE_JOYSTICKS = "SDL_AUTO_UPDATE_JOYSTICKS"
SDL_HINT_AUTO_UPDATE_SENSORS = "SDL_AUTO_UPDATE_SENSORS"
SDL_HINT_BMP_SAVE_LEGACY_FORMAT = "SDL_BMP_SAVE_LEGACY_FORMAT"
SDL_HINT_CAMERA_DRIVER = "SDL_CAMERA_DRIVER"
SDL_HINT_CPU_FEATURE_MASK = "SDL_CPU_FEATURE_MASK"
SDL_HINT_JOYSTICK_DIRECTINPUT = "SDL_JOYSTICK_DIRECTINPUT"
SDL_HINT_FILE_DIALOG_DRIVER = "SDL_FILE_DIALOG_DRIVER"
SDL_HINT_DISPLAY_USABLE_BOUNDS = "SDL_DISPLAY_USABLE_BOUNDS"
SDL_HINT_EMSCRIPTEN_ASYNCIFY = "SDL_EMSCRIPTEN_ASYNCIFY"
SDL_HINT_EMSCRIPTEN_CANVAS_SELECTOR = "SDL_EMSCRIPTEN_CANVAS_SELECTOR"
SDL_HINT_EMSCRIPTEN_KEYBOARD_ELEMENT = "SDL_EMSCRIPTEN_KEYBOARD_ELEMENT"
SDL_HINT_ENABLE_SCREEN_KEYBOARD = "SDL_ENABLE_SCREEN_KEYBOARD"
SDL_HINT_EVDEV_DEVICES = "SDL_EVDEV_DEVICES"
SDL_HINT_EVENT_LOGGING = "SDL_EVENT_LOGGING"
SDL_HINT_FORCE_RAISEWINDOW = "SDL_FORCE_RAISEWINDOW"
SDL_HINT_FRAMEBUFFER_ACCELERATION = "SDL_FRAMEBUFFER_ACCELERATION"
SDL_HINT_GAMECONTROLLERCONFIG = "SDL_GAMECONTROLLERCONFIG"
SDL_HINT_GAMECONTROLLERCONFIG_FILE = "SDL_GAMECONTROLLERCONFIG_FILE"
SDL_HINT_GAMECONTROLLERTYPE = "SDL_GAMECONTROLLERTYPE"
SDL_HINT_GAMECONTROLLER_IGNORE_DEVICES = "SDL_GAMECONTROLLER_IGNORE_DEVICES"
SDL_HINT_GAMECONTROLLER_IGNORE_DEVICES_EXCEPT = "SDL_GAMECONTROLLER_IGNORE_DEVICES_EXCEPT"
SDL_HINT_GAMECONTROLLER_SENSOR_FUSION = "SDL_GAMECONTROLLER_SENSOR_FUSION"
SDL_HINT_GDK_TEXTINPUT_DEFAULT_TEXT = "SDL_GDK_TEXTINPUT_DEFAULT_TEXT"
SDL_HINT_GDK_TEXTINPUT_DESCRIPTION = "SDL_GDK_TEXTINPUT_DESCRIPTION"
SDL_HINT_GDK_TEXTINPUT_MAX_LENGTH = "SDL_GDK_TEXTINPUT_MAX_LENGTH"
SDL_HINT_GDK_TEXTINPUT_SCOPE = "SDL_GDK_TEXTINPUT_SCOPE"
SDL_HINT_GDK_TEXTINPUT_TITLE = "SDL_GDK_TEXTINPUT_TITLE"
SDL_HINT_HIDAPI_LIBUSB = "SDL_HIDAPI_LIBUSB"
SDL_HINT_HIDAPI_LIBUSB_WHITELIST = "SDL_HIDAPI_LIBUSB_WHITELIST"
SDL_HINT_HIDAPI_UDEV = "SDL_HIDAPI_UDEV"
SDL_HINT_GPU_DRIVER = "SDL_GPU_DRIVER"
SDL_HINT_HIDAPI_ENUMERATE_ONLY_CONTROLLERS = "SDL_HIDAPI_ENUMERATE_ONLY_CONTROLLERS"
SDL_HINT_HIDAPI_IGNORE_DEVICES = "SDL_HIDAPI_IGNORE_DEVICES"
SDL_HINT_IME_IMPLEMENTED_UI = "SDL_IME_IMPLEMENTED_UI"
SDL_HINT_IOS_HIDE_HOME_INDICATOR = "SDL_IOS_HIDE_HOME_INDICATOR"
SDL_HINT_JOYSTICK_ALLOW_BACKGROUND_EVENTS = "SDL_JOYSTICK_ALLOW_BACKGROUND_EVENTS"
SDL_HINT_JOYSTICK_ARCADESTICK_DEVICES = "SDL_JOYSTICK_ARCADESTICK_DEVICES"
SDL_HINT_JOYSTICK_ARCADESTICK_DEVICES_EXCLUDED = "SDL_JOYSTICK_ARCADESTICK_DEVICES_EXCLUDED"
SDL_HINT_JOYSTICK_BLACKLIST_DEVICES = "SDL_JOYSTICK_BLACKLIST_DEVICES"
SDL_HINT_JOYSTICK_BLACKLIST_DEVICES_EXCLUDED = "SDL_JOYSTICK_BLACKLIST_DEVICES_EXCLUDED"
SDL_HINT_JOYSTICK_DEVICE = "SDL_JOYSTICK_DEVICE"
SDL_HINT_JOYSTICK_ENHANCED_REPORTS = "SDL_JOYSTICK_ENHANCED_REPORTS"
SDL_HINT_JOYSTICK_FLIGHTSTICK_DEVICES = "SDL_JOYSTICK_FLIGHTSTICK_DEVICES"
SDL_HINT_JOYSTICK_FLIGHTSTICK_DEVICES_EXCLUDED = "SDL_JOYSTICK_FLIGHTSTICK_DEVICES_EXCLUDED"
SDL_HINT_JOYSTICK_GAMEINPUT = "SDL_JOYSTICK_GAMEINPUT"
SDL_HINT_JOYSTICK_GAMECUBE_DEVICES = "SDL_JOYSTICK_GAMECUBE_DEVICES"
SDL_HINT_JOYSTICK_GAMECUBE_DEVICES_EXCLUDED = "SDL_JOYSTICK_GAMECUBE_DEVICES_EXCLUDED"
SDL_HINT_JOYSTICK_HIDAPI = "SDL_JOYSTICK_HIDAPI"
SDL_HINT_JOYSTICK_HIDAPI_COMBINE_JOY_CONS = "SDL_JOYSTICK_HIDAPI_COMBINE_JOY_CONS"
SDL_HINT_JOYSTICK_HIDAPI_GAMECUBE = "SDL_JOYSTICK_HIDAPI_GAMECUBE"
SDL_HINT_JOYSTICK_HIDAPI_GAMECUBE_RUMBLE_BRAKE = "SDL_JOYSTICK_HIDAPI_GAMECUBE_RUMBLE_BRAKE"
SDL_HINT_JOYSTICK_HIDAPI_JOY_CONS = "SDL_JOYSTICK_HIDAPI_JOY_CONS"
SDL_HINT_JOYSTICK_HIDAPI_JOYCON_HOME_LED = "SDL_JOYSTICK_HIDAPI_JOYCON_HOME_LED"
SDL_HINT_JOYSTICK_HIDAPI_LUNA = "SDL_JOYSTICK_HIDAPI_LUNA"
SDL_HINT_JOYSTICK_HIDAPI_NINTENDO_CLASSIC = "SDL_JOYSTICK_HIDAPI_NINTENDO_CLASSIC"
SDL_HINT_JOYSTICK_HIDAPI_PS3 = "SDL_JOYSTICK_HIDAPI_PS3"
SDL_HINT_JOYSTICK_HIDAPI_PS3_SIXAXIS_DRIVER = "SDL_JOYSTICK_HIDAPI_PS3_SIXAXIS_DRIVER"
SDL_HINT_JOYSTICK_HIDAPI_PS4 = "SDL_JOYSTICK_HIDAPI_PS4"
SDL_HINT_JOYSTICK_HIDAPI_PS4_REPORT_INTERVAL = "SDL_JOYSTICK_HIDAPI_PS4_REPORT_INTERVAL"
SDL_HINT_JOYSTICK_HIDAPI_PS5 = "SDL_JOYSTICK_HIDAPI_PS5"
SDL_HINT_JOYSTICK_HIDAPI_PS5_PLAYER_LED = "SDL_JOYSTICK_HIDAPI_PS5_PLAYER_LED"
SDL_HINT_JOYSTICK_HIDAPI_SHIELD = "SDL_JOYSTICK_HIDAPI_SHIELD"
SDL_HINT_JOYSTICK_HIDAPI_STADIA = "SDL_JOYSTICK_HIDAPI_STADIA"
SDL_HINT_JOYSTICK_HIDAPI_STEAM = "SDL_JOYSTICK_HIDAPI_STEAM"
SDL_HINT_JOYSTICK_HIDAPI_STEAM_HOME_LED = "SDL_JOYSTICK_HIDAPI_STEAM_HOME_LED"
SDL_HINT_JOYSTICK_HIDAPI_STEAMDECK = "SDL_JOYSTICK_HIDAPI_STEAMDECK"
SDL_HINT_JOYSTICK_HIDAPI_STEAM_HORI = "SDL_JOYSTICK_HIDAPI_STEAM_HORI"
SDL_HINT_JOYSTICK_HIDAPI_SWITCH = "SDL_JOYSTICK_HIDAPI_SWITCH"
SDL_HINT_JOYSTICK_HIDAPI_SWITCH_HOME_LED = "SDL_JOYSTICK_HIDAPI_SWITCH_HOME_LED"
SDL_HINT_JOYSTICK_HIDAPI_SWITCH_PLAYER_LED = "SDL_JOYSTICK_HIDAPI_SWITCH_PLAYER_LED"
SDL_HINT_JOYSTICK_HIDAPI_VERTICAL_JOY_CONS = "SDL_JOYSTICK_HIDAPI_VERTICAL_JOY_CONS"
SDL_HINT_JOYSTICK_HIDAPI_WII = "SDL_JOYSTICK_HIDAPI_WII"
SDL_HINT_JOYSTICK_HIDAPI_WII_PLAYER_LED = "SDL_JOYSTICK_HIDAPI_WII_PLAYER_LED"
SDL_HINT_JOYSTICK_HIDAPI_XBOX = "SDL_JOYSTICK_HIDAPI_XBOX"
SDL_HINT_JOYSTICK_HIDAPI_XBOX_360 = "SDL_JOYSTICK_HIDAPI_XBOX_360"
SDL_HINT_JOYSTICK_HIDAPI_XBOX_360_PLAYER_LED = "SDL_JOYSTICK_HIDAPI_XBOX_360_PLAYER_LED"
SDL_HINT_JOYSTICK_HIDAPI_XBOX_360_WIRELESS = "SDL_JOYSTICK_HIDAPI_XBOX_360_WIRELESS"
SDL_HINT_JOYSTICK_HIDAPI_XBOX_ONE = "SDL_JOYSTICK_HIDAPI_XBOX_ONE"
SDL_HINT_JOYSTICK_HIDAPI_XBOX_ONE_HOME_LED = "SDL_JOYSTICK_HIDAPI_XBOX_ONE_HOME_LED"
SDL_HINT_JOYSTICK_IOKIT = "SDL_JOYSTICK_IOKIT"
SDL_HINT_JOYSTICK_LINUX_CLASSIC = "SDL_JOYSTICK_LINUX_CLASSIC"
SDL_HINT_JOYSTICK_LINUX_DEADZONES = "SDL_JOYSTICK_LINUX_DEADZONES"
SDL_HINT_JOYSTICK_LINUX_DIGITAL_HATS = "SDL_JOYSTICK_LINUX_DIGITAL_HATS"
SDL_HINT_JOYSTICK_LINUX_HAT_DEADZONES = "SDL_JOYSTICK_LINUX_HAT_DEADZONES"
SDL_HINT_JOYSTICK_MFI = "SDL_JOYSTICK_MFI"
SDL_HINT_JOYSTICK_RAWINPUT = "SDL_JOYSTICK_RAWINPUT"
SDL_HINT_JOYSTICK_RAWINPUT_CORRELATE_XINPUT = "SDL_JOYSTICK_RAWINPUT_CORRELATE_XINPUT"
SDL_HINT_JOYSTICK_ROG_CHAKRAM = "SDL_JOYSTICK_ROG_CHAKRAM"
SDL_HINT_JOYSTICK_THREAD = "SDL_JOYSTICK_THREAD"
SDL_HINT_JOYSTICK_THROTTLE_DEVICES = "SDL_JOYSTICK_THROTTLE_DEVICES"
SDL_HINT_JOYSTICK_THROTTLE_DEVICES_EXCLUDED = "SDL_JOYSTICK_THROTTLE_DEVICES_EXCLUDED"
SDL_HINT_JOYSTICK_WGI = "SDL_JOYSTICK_WGI"
SDL_HINT_JOYSTICK_WHEEL_DEVICES = "SDL_JOYSTICK_WHEEL_DEVICES"
SDL_HINT_JOYSTICK_WHEEL_DEVICES_EXCLUDED = "SDL_JOYSTICK_WHEEL_DEVICES_EXCLUDED"
SDL_HINT_JOYSTICK_ZERO_CENTERED_DEVICES = "SDL_JOYSTICK_ZERO_CENTERED_DEVICES"
SDL_HINT_KEYCODE_OPTIONS = "SDL_KEYCODE_OPTIONS"
SDL_HINT_KMSDRM_DEVICE_INDEX = "SDL_KMSDRM_DEVICE_INDEX"
SDL_HINT_KMSDRM_REQUIRE_DRM_MASTER = "SDL_KMSDRM_REQUIRE_DRM_MASTER"
SDL_HINT_LOGGING = "SDL_LOGGING"
SDL_HINT_MAC_BACKGROUND_APP = "SDL_MAC_BACKGROUND_APP"
SDL_HINT_MAC_CTRL_CLICK_EMULATE_RIGHT_CLICK = "SDL_MAC_CTRL_CLICK_EMULATE_RIGHT_CLICK"
SDL_HINT_MAC_OPENGL_ASYNC_DISPATCH = "SDL_MAC_OPENGL_ASYNC_DISPATCH"
SDL_HINT_MAC_OPTION_AS_ALT = "SDL_MAC_OPTION_AS_ALT"
SDL_HINT_MAC_SCROLL_MOMENTUM = "SDL_MAC_SCROLL_MOMENTUM"
SDL_HINT_MAIN_CALLBACK_RATE = "SDL_MAIN_CALLBACK_RATE"
SDL_HINT_MOUSE_AUTO_CAPTURE = "SDL_MOUSE_AUTO_CAPTURE"
SDL_HINT_MOUSE_DOUBLE_CLICK_RADIUS = "SDL_MOUSE_DOUBLE_CLICK_RADIUS"
SDL_HINT_MOUSE_DOUBLE_CLICK_TIME = "SDL_MOUSE_DOUBLE_CLICK_TIME"
SDL_HINT_MOUSE_DEFAULT_SYSTEM_CURSOR = "SDL_MOUSE_DEFAULT_SYSTEM_CURSOR"
SDL_HINT_MOUSE_EMULATE_WARP_WITH_RELATIVE = "SDL_MOUSE_EMULATE_WARP_WITH_RELATIVE"
SDL_HINT_MOUSE_FOCUS_CLICKTHROUGH = "SDL_MOUSE_FOCUS_CLICKTHROUGH"
SDL_HINT_MOUSE_NORMAL_SPEED_SCALE = "SDL_MOUSE_NORMAL_SPEED_SCALE"
SDL_HINT_MOUSE_RELATIVE_MODE_CENTER = "SDL_MOUSE_RELATIVE_MODE_CENTER"
SDL_HINT_MOUSE_RELATIVE_SPEED_SCALE = "SDL_MOUSE_RELATIVE_SPEED_SCALE"
SDL_HINT_MOUSE_RELATIVE_SYSTEM_SCALE = "SDL_MOUSE_RELATIVE_SYSTEM_SCALE"
SDL_HINT_MOUSE_RELATIVE_WARP_MOTION = "SDL_MOUSE_RELATIVE_WARP_MOTION"
SDL_HINT_MOUSE_RELATIVE_CURSOR_VISIBLE = "SDL_MOUSE_RELATIVE_CURSOR_VISIBLE"
SDL_HINT_MOUSE_TOUCH_EVENTS = "SDL_MOUSE_TOUCH_EVENTS"
SDL_HINT_MUTE_CONSOLE_KEYBOARD = "SDL_MUTE_CONSOLE_KEYBOARD"
SDL_HINT_NO_SIGNAL_HANDLERS = "SDL_NO_SIGNAL_HANDLERS"
SDL_HINT_OPENGL_LIBRARY = "SDL_OPENGL_LIBRARY"
SDL_HINT_EGL_LIBRARY = "SDL_OPENGL_LIBRARY"
SDL_HINT_OPENGL_ES_DRIVER = "SDL_OPENGL_ES_DRIVER"
SDL_HINT_OPENVR_LIBRARY = "SDL_OPENVR_LIBRARY"
SDL_HINT_ORIENTATIONS = "SDL_ORIENTATIONS"
SDL_HINT_PEN_DELAY_MOUSE_BUTTON = "SDL_PEN_DELAY_MOUSE_BUTTON"
SDL_HINT_PEN_NOT_MOUSE = "SDL_PEN_NOT_MOUSE"
SDL_HINT_POLL_SENTINEL = "SDL_POLL_SENTINEL"
SDL_HINT_PREFERRED_LOCALES = "SDL_PREFERRED_LOCALES"
SDL_HINT_QUIT_ON_LAST_WINDOW_CLOSE = "SDL_QUIT_ON_LAST_WINDOW_CLOSE"
SDL_HINT_RENDER_DIRECT3D_THREADSAFE = "SDL_RENDER_DIRECT3D_THREADSAFE"
SDL_HINT_RENDER_DIRECT3D11_DEBUG = "SDL_RENDER_DIRECT3D11_DEBUG"
SDL_HINT_RENDER_VULKAN_DEBUG = "SDL_RENDER_VULKAN_DEBUG"
SDL_HINT_RENDER_GPU_DEBUG = "SDL_RENDER_GPU_DEBUG"
SDL_HINT_RENDER_GPU_LOW_POWER = "SDL_RENDER_GPU_LOW_POWER"
SDL_HINT_RENDER_DRIVER = "SDL_RENDER_DRIVER"
SDL_HINT_RENDER_LINE_METHOD = "SDL_RENDER_LINE_METHOD"
SDL_HINT_RENDER_METAL_PREFER_LOW_POWER_DEVICE = "SDL_RENDER_METAL_PREFER_LOW_POWER_DEVICE"
SDL_HINT_RENDER_VSYNC = "SDL_RENDER_VSYNC"
SDL_HINT_RETURN_KEY_HIDES_IME = "SDL_RETURN_KEY_HIDES_IME"
SDL_HINT_ROG_GAMEPAD_MICE = "SDL_ROG_GAMEPAD_MICE"
SDL_HINT_ROG_GAMEPAD_MICE_EXCLUDED = "SDL_ROG_GAMEPAD_MICE_EXCLUDED"
SDL_HINT_RPI_VIDEO_LAYER = "SDL_RPI_VIDEO_LAYER"
SDL_HINT_SCREENSAVER_INHIBIT_ACTIVITY_NAME = "SDL_SCREENSAVER_INHIBIT_ACTIVITY_NAME"
SDL_HINT_SHUTDOWN_DBUS_ON_QUIT = "SDL_SHUTDOWN_DBUS_ON_QUIT"
SDL_HINT_STORAGE_TITLE_DRIVER = "SDL_STORAGE_TITLE_DRIVER"
SDL_HINT_STORAGE_USER_DRIVER = "SDL_STORAGE_USER_DRIVER"
SDL_HINT_THREAD_FORCE_REALTIME_TIME_CRITICAL = "SDL_THREAD_FORCE_REALTIME_TIME_CRITICAL"
SDL_HINT_THREAD_PRIORITY_POLICY = "SDL_THREAD_PRIORITY_POLICY"
SDL_HINT_TIMER_RESOLUTION = "SDL_TIMER_RESOLUTION"
SDL_HINT_TOUCH_MOUSE_EVENTS = "SDL_TOUCH_MOUSE_EVENTS"
SDL_HINT_TRACKPAD_IS_TOUCH_ONLY = "SDL_TRACKPAD_IS_TOUCH_ONLY"
SDL_HINT_TV_REMOTE_AS_JOYSTICK = "SDL_TV_REMOTE_AS_JOYSTICK"
SDL_HINT_VIDEO_ALLOW_SCREENSAVER = "SDL_VIDEO_ALLOW_SCREENSAVER"
SDL_HINT_VIDEO_DISPLAY_PRIORITY = "SDL_VIDEO_DISPLAY_PRIORITY"
SDL_HINT_VIDEO_DOUBLE_BUFFER = "SDL_VIDEO_DOUBLE_BUFFER"
SDL_HINT_VIDEO_DRIVER = "SDL_VIDEO_DRIVER"
SDL_HINT_VIDEO_DUMMY_SAVE_FRAMES = "SDL_VIDEO_DUMMY_SAVE_FRAMES"
SDL_HINT_VIDEO_EGL_ALLOW_GETDISPLAY_FALLBACK = "SDL_VIDEO_EGL_ALLOW_GETDISPLAY_FALLBACK"
SDL_HINT_VIDEO_FORCE_EGL = "SDL_VIDEO_FORCE_EGL"
SDL_HINT_VIDEO_MAC_FULLSCREEN_SPACES = "SDL_VIDEO_MAC_FULLSCREEN_SPACES"
SDL_HINT_VIDEO_MAC_FULLSCREEN_MENU_VISIBILITY = "SDL_VIDEO_MAC_FULLSCREEN_MENU_VISIBILITY"
SDL_HINT_VIDEO_MINIMIZE_ON_FOCUS_LOSS = "SDL_VIDEO_MINIMIZE_ON_FOCUS_LOSS"
SDL_HINT_VIDEO_OFFSCREEN_SAVE_FRAMES = "SDL_VIDEO_OFFSCREEN_SAVE_FRAMES"
SDL_HINT_VIDEO_SYNC_WINDOW_OPERATIONS = "SDL_VIDEO_SYNC_WINDOW_OPERATIONS"
SDL_HINT_VIDEO_WAYLAND_ALLOW_LIBDECOR = "SDL_VIDEO_WAYLAND_ALLOW_LIBDECOR"
SDL_HINT_VIDEO_WAYLAND_MODE_EMULATION = "SDL_VIDEO_WAYLAND_MODE_EMULATION"
SDL_HINT_VIDEO_WAYLAND_MODE_SCALING = "SDL_VIDEO_WAYLAND_MODE_SCALING"
SDL_HINT_VIDEO_WAYLAND_PREFER_LIBDECOR = "SDL_VIDEO_WAYLAND_PREFER_LIBDECOR"
SDL_HINT_VIDEO_WAYLAND_SCALE_TO_DISPLAY = "SDL_VIDEO_WAYLAND_SCALE_TO_DISPLAY"
SDL_HINT_VIDEO_WIN_D3DCOMPILER = "SDL_VIDEO_WIN_D3DCOMPILER"
SDL_HINT_VIDEO_X11_NET_WM_BYPASS_COMPOSITOR = "SDL_VIDEO_X11_NET_WM_BYPASS_COMPOSITOR"
SDL_HINT_VIDEO_X11_NET_WM_PING = "SDL_VIDEO_X11_NET_WM_PING"
SDL_HINT_VIDEO_X11_NODIRECTCOLOR = "SDL_VIDEO_X11_NODIRECTCOLOR"
SDL_HINT_VIDEO_X11_SCALING_FACTOR = "SDL_VIDEO_X11_SCALING_FACTOR"
SDL_HINT_VIDEO_X11_VISUALID = "SDL_VIDEO_X11_VISUALID"
SDL_HINT_VIDEO_X11_WINDOW_VISUALID = "SDL_VIDEO_X11_WINDOW_VISUALID"
SDL_HINT_VIDEO_X11_XRANDR = "SDL_VIDEO_X11_XRANDR"
SDL_HINT_VITA_ENABLE_BACK_TOUCH = "SDL_VITA_ENABLE_BACK_TOUCH"
SDL_HINT_VITA_ENABLE_FRONT_TOUCH = "SDL_VITA_ENABLE_FRONT_TOUCH"
SDL_HINT_VITA_MODULE_PATH = "SDL_VITA_MODULE_PATH"
SDL_HINT_VITA_PVR_INIT = "SDL_VITA_PVR_INIT"
SDL_HINT_VITA_RESOLUTION = "SDL_VITA_RESOLUTION"
SDL_HINT_VITA_PVR_OPENGL = "SDL_VITA_PVR_OPENGL"
SDL_HINT_VITA_TOUCH_MOUSE_DEVICE = "SDL_VITA_TOUCH_MOUSE_DEVICE"
SDL_HINT_VULKAN_DISPLAY = "SDL_VULKAN_DISPLAY"
SDL_HINT_VULKAN_LIBRARY = "SDL_VULKAN_LIBRARY"
SDL_HINT_WAVE_FACT_CHUNK = "SDL_WAVE_FACT_CHUNK"
SDL_HINT_WAVE_CHUNK_LIMIT = "SDL_WAVE_CHUNK_LIMIT"
SDL_HINT_WAVE_RIFF_CHUNK_SIZE = "SDL_WAVE_RIFF_CHUNK_SIZE"
SDL_HINT_WAVE_TRUNCATION = "SDL_WAVE_TRUNCATION"
SDL_HINT_WINDOW_ACTIVATE_WHEN_RAISED = "SDL_WINDOW_ACTIVATE_WHEN_RAISED"
SDL_HINT_WINDOW_ACTIVATE_WHEN_SHOWN = "SDL_WINDOW_ACTIVATE_WHEN_SHOWN"
SDL_HINT_WINDOW_ALLOW_TOPMOST = "SDL_WINDOW_ALLOW_TOPMOST"
SDL_HINT_WINDOW_FRAME_USABLE_WHILE_CURSOR_HIDDEN = "SDL_WINDOW_FRAME_USABLE_WHILE_CURSOR_HIDDEN"
SDL_HINT_WINDOWS_CLOSE_ON_ALT_F4 = "SDL_WINDOWS_CLOSE_ON_ALT_F4"
SDL_HINT_WINDOWS_ENABLE_MENU_MNEMONICS = "SDL_WINDOWS_ENABLE_MENU_MNEMONICS"
SDL_HINT_WINDOWS_ENABLE_MESSAGELOOP = "SDL_WINDOWS_ENABLE_MESSAGELOOP"
SDL_HINT_WINDOWS_GAMEINPUT = "SDL_WINDOWS_GAMEINPUT"
SDL_HINT_WINDOWS_RAW_KEYBOARD = "SDL_WINDOWS_RAW_KEYBOARD"
SDL_HINT_WINDOWS_FORCE_SEMAPHORE_KERNEL = "SDL_WINDOWS_FORCE_SEMAPHORE_KERNEL"
SDL_HINT_WINDOWS_INTRESOURCE_ICON = "SDL_WINDOWS_INTRESOURCE_ICON"
SDL_HINT_WINDOWS_INTRESOURCE_ICON_SMALL = "SDL_WINDOWS_INTRESOURCE_ICON_SMALL"
SDL_HINT_WINDOWS_USE_D3D9EX = "SDL_WINDOWS_USE_D3D9EX"
SDL_HINT_WINDOWS_ERASE_BACKGROUND_MODE = "SDL_WINDOWS_ERASE_BACKGROUND_MODE"
SDL_HINT_X11_FORCE_OVERRIDE_REDIRECT = "SDL_X11_FORCE_OVERRIDE_REDIRECT"
SDL_HINT_X11_WINDOW_TYPE = "SDL_X11_WINDOW_TYPE"
SDL_HINT_X11_XCB_LIBRARY = "SDL_X11_XCB_LIBRARY"
SDL_HINT_XINPUT_ENABLED = "SDL_XINPUT_ENABLED"
SDL_HINT_ASSERT = "SDL_ASSERT"
SDL_HINT_PEN_MOUSE_EVENTS = "SDL_PEN_MOUSE_EVENTS"
SDL_HINT_PEN_TOUCH_EVENTS = "SDL_PEN_TOUCH_EVENTS"

SDL_HintPriority = ctypes.c_int

SDL_HINT_DEFAULT = 0
SDL_HINT_NORMAL = 1
SDL_HINT_OVERRIDE = 2

SDL_FUNC("SDL_SetHintWithPriority", ctypes.c_bool, ctypes.c_char_p, ctypes.c_char_p, SDL_HintPriority)
SDL_FUNC("SDL_SetHint", ctypes.c_bool, ctypes.c_char_p, ctypes.c_char_p)
SDL_FUNC("SDL_ResetHint", ctypes.c_bool, ctypes.c_char_p)
SDL_FUNC("SDL_ResetHints", None)
SDL_FUNC("SDL_GetHint", ctypes.c_char_p, ctypes.c_char_p)
SDL_FUNC("SDL_GetHintBoolean", ctypes.c_bool, ctypes.c_char_p, ctypes.c_bool)

SDL_HintCallback = ctypes.CFUNCTYPE(None, ctypes.c_void_p, ctypes.c_char_p, ctypes.c_char_p, ctypes.c_char_p)

SDL_FUNC("SDL_AddHintCallback", ctypes.c_bool, ctypes.c_char_p, SDL_HintCallback, ctypes.c_void_p)
SDL_FUNC("SDL_RemoveHintCallback", None, ctypes.c_char_p, SDL_HintCallback, ctypes.c_void_p)


from .__init__ import sys, ctypes, array, \
    SDL_FUNC, SDL_SET_CURRENT_BINARY, SDL_BINARY

SDL_SET_CURRENT_BINARY(SDL_BINARY)

SDL_LIL_ENDIAN, SDL_BIG_ENDIAN = 1234, 4321
SDL_BYTEORDER = SDL_LIL_ENDIAN \
    if sys.byteorder == "little" else SDL_BIG_ENDIAN

sys.byteorder

def SDL_Swap16(x):
    return ((x << 8) & 0xFF00) | ((x >> 8) & 0x00FF)

def SDL_Swap32(x):
    return ((x << 24) & 0xFF000000) | ((x << 8) & 0x00FF0000) | ((x >> 8) & 0x0000FF00) | ((x >> 24) & 0x000000FF)

def SDL_Swap64(x):
    return (SDL_Swap32(x & 0xFFFFFFFF) << 32) | (SDL_Swap32(x >> 32 & 0xFFFFFFFF))

def SDL_SwapFloat(x):
    ar = array.array("d", (x, ))
    ar.byteswap()
    return ar[0]

if SDL_BYTEORDER == SDL_LIL_ENDIAN:
    SDL_SwapLE16 = lambda x: x
    SDL_SwapLE32 = lambda x: x
    SDL_SwapLE64 = lambda x: x
    SDL_SwapFloatLE = lambda x: x
    SDL_SwapBE16 = SDL_Swap16
    SDL_SwapBE32 = SDL_Swap32
    SDL_SwapBE64 = SDL_Swap64
    SDL_SwapFloatBE = SDL_SwapFloat

else:
    SDL_SwapLE16 = SDL_Swap16
    SDL_SwapLE32 = SDL_Swap32
    SDL_SwapLE64 = SDL_Swap64
    SDL_SwapFloatLE = SDL_SwapFloat
    SDL_SwapBE16 = lambda x: x
    SDL_SwapBE32 = lambda x: x
    SDL_SwapBE64 = lambda x: x
    SDL_SwapFloatBE = lambda x: x


from .__init__ import ctypes, \
    SDL_FUNC, SDL_SET_CURRENT_BINARY, SDL_GET_BINARY, SDL_BINARY

SDL_SET_CURRENT_BINARY(SDL_BINARY)

SDL_FUNC("SDL_SetError", ctypes.c_bool, ctypes.c_char_p)
SDL_FUNC("SDL_SetErrorV", ctypes.c_bool, ctypes.c_char_p, ctypes.c_void_p)
SDL_FUNC("SDL_OutOfMemory", ctypes.c_bool)
SDL_FUNC("SDL_GetError", ctypes.c_char_p)
SDL_FUNC("SDL_ClearError", ctypes.c_bool)

def SDL_Unsupported() -> ctypes.c_int:
    return SDL_GET_BINARY(SDL_BINARY).SDL_SetError("That operation is not supported".encode())

def SDL_InvalidParamError(param: ctypes.c_char_p) -> ctypes.c_int:
    if isinstance(param, bytes): param = param.decode()
    return SDL_GET_BINARY(SDL_BINARY).SDL_SetError(f"Parameter '{param}' is invalid".encode())


from .__init__ import ctypes, \
    SDL_FUNC, SDL_SET_CURRENT_BINARY, SDL_BINARY

from .SDL_mouse import SDL_MouseID

SDL_SET_CURRENT_BINARY(SDL_BINARY)

SDL_TouchID = ctypes.c_uint64
SDL_FingerID = ctypes.c_uint64

SDL_TouchDeviceType = ctypes.c_int

SDL_TOUCH_DEVICE_INVALID = -1
SDL_TOUCH_DEVICE_DIRECT = 0
SDL_TOUCH_DEVICE_INDIRECT_ABSOLUTE = 1
SDL_TOUCH_DEVICE_INDIRECT_RELATIVE = 2

class SDL_Finger(ctypes.Structure):
    _fields_ = [
        ("id", SDL_FingerID),
        ("x", ctypes.c_float),
        ("y", ctypes.c_float),
        ("pressure", ctypes.c_float)
    ]

SDL_TOUCH_MOUSEID = SDL_MouseID(-1)
SDL_MOUSE_TOUCHID = SDL_TouchID(-1)

SDL_FUNC("SDL_GetTouchDevices", ctypes.POINTER(SDL_TouchID), ctypes.POINTER(ctypes.c_int))
SDL_FUNC("SDL_GetTouchDeviceName", ctypes.c_char_p, SDL_TouchID)
SDL_FUNC("SDL_GetTouchDeviceType", SDL_TouchDeviceType, SDL_TouchID)
SDL_FUNC("SDL_GetTouchFingers", ctypes.POINTER(ctypes.POINTER(SDL_Finger)), SDL_TouchID, ctypes.POINTER(ctypes.c_int))



from .__init__ import ctypes, SDL_SYSTEM, \
    SDL_FUNC, SDL_SET_CURRENT_BINARY, SDL_BINARY

from .SDL_video import SDL_DisplayID

SDL_SET_CURRENT_BINARY(SDL_BINARY)

if SDL_SYSTEM in ["Windows"]:
    import ctypes.wintypes as wintypes

    MSG = wintypes.tagMSG
    SDL_WindowsMessageHook = ctypes.CFUNCTYPE(ctypes.c_bool, ctypes.c_void_p, ctypes.POINTER(MSG))
    SDL_FUNC("SDL_SetWindowsMessageHook", None, SDL_WindowsMessageHook, ctypes.c_void_p)

    SDL_FUNC("SDL_GetDirect3D9AdapterIndex", ctypes.c_int, SDL_DisplayID)
    SDL_FUNC("SDL_GetDXGIOutputInfo", ctypes.c_bool, SDL_DisplayID, ctypes.POINTER(ctypes.c_int), ctypes.POINTER(ctypes.c_int))

    class XEvent(ctypes.Union):
        ...

    SDL_X11EventHook = ctypes.CFUNCTYPE(ctypes.c_bool, ctypes.c_void_p, ctypes.POINTER(XEvent))
    SDL_FUNC("SDL_SetX11EventHook", None, SDL_X11EventHook, ctypes.c_void_p) 

if SDL_SYSTEM in ["Linux"]:
    SDL_FUNC("SDL_SetLinuxThreadPriority", ctypes.c_bool, ctypes.c_int64, ctypes.c_int)
    SDL_FUNC("SDL_SetLinuxThreadPriorityAndPolicy", ctypes.c_bool, ctypes.c_int64, ctypes.c_int, ctypes.c_int)

SDL_FUNC("SDL_IsTablet", ctypes.c_bool)
SDL_FUNC("SDL_IsTV", ctypes.c_bool)

SDL_Sandbox = ctypes.c_int

SDL_SANDBOX_NONE = 0
SDL_SANDBOX_UNKNOWN_CONTAINER = 1
SDL_SANDBOX_FLATPAK = 2
SDL_SANDBOX_SNAP = 3
SDL_SANDBOX_MACOS = 4

SDL_FUNC("SDL_GetSandbox", SDL_Sandbox)

SDL_FUNC("SDL_OnApplicationWillTerminate", None)
SDL_FUNC("SDL_OnApplicationDidReceiveMemoryWarning", None)
SDL_FUNC("SDL_OnApplicationWillEnterBackground", None)
SDL_FUNC("SDL_OnApplicationDidEnterBackground", None)
SDL_FUNC("SDL_OnApplicationWillEnterForeground", None)
SDL_FUNC("SDL_OnApplicationDidEnterForeground", None)


from .__init__ import ctypes, \
    SDL_FUNC, SDL_SET_CURRENT_BINARY, SDL_BINARY

from .SDL_video import SDL_Window

SDL_SET_CURRENT_BINARY(SDL_BINARY)

SDL_MessageBoxFlags = ctypes.c_uint32

SDL_MESSAGEBOX_ERROR = 0x00000010
SDL_MESSAGEBOX_WARNING = 0x00000020
SDL_MESSAGEBOX_INFORMATION = 0x00000040
SDL_MESSAGEBOX_BUTTONS_LEFT_TO_RIGHT = 0x00000080
SDL_MESSAGEBOX_BUTTONS_RIGHT_TO_LEFT = 0x00000100

SDL_MessageBoxButtonFlags = ctypes.c_uint32

SDL_MESSAGEBOX_BUTTON_RETURNKEY_DEFAULT = 0x00000001
SDL_MESSAGEBOX_BUTTON_ESCAPEKEY_DEFAULT = 0x00000002

class SDL_MessageBoxButtonData(ctypes.Structure):
    _fields_ = [
        ("flags", SDL_MessageBoxButtonFlags),
        ("buttonID", ctypes.c_int),
        ("text", ctypes.c_char_p)
    ]

class SDL_MessageBoxColor(ctypes.Structure):
    _fields_ = [
        ("r", ctypes.c_uint8),
        ("g", ctypes.c_uint8),
        ("b", ctypes.c_uint8)
    ]

SDL_MessageBoxColorType = ctypes.c_int

SDL_MESSAGEBOX_COLOR_BACKGROUND = 0
SDL_MESSAGEBOX_COLOR_TEXT = 1
SDL_MESSAGEBOX_COLOR_BUTTON_BORDER = 2
SDL_MESSAGEBOX_COLOR_BUTTON_BACKGROUND = 3
SDL_MESSAGEBOX_COLOR_BUTTON_SELECTED = 4
SDL_MESSAGEBOX_COLOR_COUNT = 5

class SDL_MessageBoxColorScheme(ctypes.Structure):
    _fields_ = [
        ("colors", SDL_MessageBoxColor * SDL_MESSAGEBOX_COLOR_COUNT)
    ]

class SDL_MessageBoxData(ctypes.Structure):
    _fields_ = [
        ("flags", SDL_MessageBoxFlags),
        ("window", ctypes.POINTER(SDL_Window)),
        ("title", ctypes.c_char_p),
        ("message", ctypes.c_char_p),
        ("numbuttons", ctypes.c_int),
        ("buttons", ctypes.POINTER(SDL_MessageBoxButtonData)),
        ("colorScheme", ctypes.POINTER(SDL_MessageBoxColorScheme))
    ]

SDL_FUNC("SDL_ShowMessageBox", ctypes.c_bool, ctypes.POINTER(SDL_MessageBoxData), ctypes.POINTER(ctypes.c_int))
SDL_FUNC("SDL_ShowSimpleMessageBox", ctypes.c_bool, SDL_MessageBoxFlags, ctypes.c_char_p, ctypes.c_char_p, ctypes.POINTER(SDL_Window))


from .__init__ import ctypes, \
    SDL_FUNC, SDL_SET_CURRENT_BINARY, SDL_BINARY

from .SDL_endian import SDL_BYTEORDER, SDL_BIG_ENDIAN

SDL_SET_CURRENT_BINARY(SDL_BINARY)

SDL_ALPHA_OPAQUE = 255
SDL_ALPHA_OPAQUE_FLOAT = 1.0

SDL_ALPHA_TRANSPARENT = 0
SDL_ALPHA_TRANSPARENT_FLOAT = 0.0

SDL_PixelType = ctypes.c_int

SDL_PIXELTYPE_UNKNOWN = 0
SDL_PIXELTYPE_INDEX1 = 1
SDL_PIXELTYPE_INDEX4 = 2
SDL_PIXELTYPE_INDEX8 = 3
SDL_PIXELTYPE_PACKED8 = 4
SDL_PIXELTYPE_PACKED16 = 5
SDL_PIXELTYPE_PACKED32 = 6
SDL_PIXELTYPE_ARRAYU8 = 7
SDL_PIXELTYPE_ARRAYU16 = 8
SDL_PIXELTYPE_ARRAYU32 = 9
SDL_PIXELTYPE_ARRAYF16 = 10
SDL_PIXELTYPE_ARRAYF32 = 11
SDL_PIXELTYPE_INDEX2 = 12

SDL_BitmapOrder = ctypes.c_int

SDL_BITMAPORDER_NONE = 0
SDL_BITMAPORDER_4321 = 1
SDL_BITMAPORDER_1234 = 2

SDL_PackedOrder = ctypes.c_int

SDL_PACKEDORDER_NONE = 0
SDL_PACKEDORDER_XRGB = 1
SDL_PACKEDORDER_RGBX = 2
SDL_PACKEDORDER_ARGB = 3
SDL_PACKEDORDER_RGBA = 4
SDL_PACKEDORDER_XBGR = 5
SDL_PACKEDORDER_BGRX = 6
SDL_PACKEDORDER_ABGR = 7
SDL_PACKEDORDER_BGRA = 8

SDL_ArrayOrder = ctypes.c_int

SDL_ARRAYORDER_NONE = 0
SDL_ARRAYORDER_RGB = 1
SDL_ARRAYORDER_RGBA = 2
SDL_ARRAYORDER_ARGB = 3
SDL_ARRAYORDER_BGR = 4
SDL_ARRAYORDER_BGRA = 5
SDL_ARRAYORDER_ABGR = 6

SDL_PackedLayout = ctypes.c_int

SDL_PACKEDLAYOUT_NONE = 0
SDL_PACKEDLAYOUT_332 = 1
SDL_PACKEDLAYOUT_4444 = 2
SDL_PACKEDLAYOUT_1555 = 3
SDL_PACKEDLAYOUT_5551 = 4
SDL_PACKEDLAYOUT_565 = 5
SDL_PACKEDLAYOUT_8888 = 6
SDL_PACKEDLAYOUT_2101010 = 7
SDL_PACKEDLAYOUT_1010102 = 8

def SDL_FOURCC(a, b, c, d):
    return (ord(a) << 0) | (ord(b) << 8) | (ord(c) << 16) | (ord(d) << 24)

def SDL_DEFINE_PIXELFORMAT(ptype, order, layout, bits, pbytes):
    return (1 << 28) | (ptype << 24) | (order << 20) | (layout << 16) | (bits << 8) | (pbytes << 0)

SDL_DEFINE_PIXELFOURCC = SDL_FOURCC
SDL_ISPIXELFORMAT_FOURCC = lambda format: format and SDL_PIXELFLAG(format) != 1

SDL_PIXELFLAG = lambda x: (x >> 28) & 0x0F
SDL_PIXELTYPE = lambda x: (x >> 24) & 0x0F
SDL_PIXELORDER = lambda x: (x >> 20) & 0x0F
SDL_PIXELLAYOUT = lambda x: (x >> 16) & 0x0F
SDL_BITSPERPIXEL = lambda x: (x >> 8) & 0xFF

def SDL_BYTESPERPIXEL(x):
    if SDL_ISPIXELFORMAT_FOURCC(x):
        return 2 if x in (SDL_PIXELFORMAT_YUY2, SDL_PIXELFORMAT_UYVY, SDL_PIXELFORMAT_YVYU, SDL_PIXELFORMAT_P010) else 1
        
    else:
        return (x >> 0) & 0xFF

def SDL_ISPIXELFORMAT_INDEXED(format):
    return not SDL_ISPIXELFORMAT_FOURCC(format) \
        and (SDL_PIXELTYPE(format) in [SDL_PIXELTYPE_INDEX1, SDL_PIXELTYPE_INDEX2, SDL_PIXELTYPE_INDEX4, SDL_PIXELTYPE_INDEX8])

def SDL_ISPIXELFORMAT_PACKED(format):
    return not SDL_ISPIXELFORMAT_FOURCC(format) \
        and (SDL_PIXELTYPE(format) in [SDL_PIXELTYPE_PACKED8, SDL_PIXELTYPE_PACKED16, SDL_PIXELTYPE_PACKED32])

def SDL_ISPIXELFORMAT_ARRAY(format):
    return not SDL_ISPIXELFORMAT_FOURCC(format) \
        and SDL_PIXELTYPE(format) in [SDL_PIXELTYPE_ARRAYU8, SDL_PIXELTYPE_ARRAYU16, SDL_PIXELTYPE_ARRAYU32, SDL_PIXELTYPE_ARRAYF16, SDL_PIXELTYPE_ARRAYF32]

def SDL_ISPIXELFORMAT_10BIT(format):
    return not SDL_ISPIXELFORMAT_FOURCC(format) and (SDL_PIXELTYPE(format) == SDL_PIXELTYPE_PACKED32 and SDL_PIXELLAYOUT(format) == SDL_PACKEDLAYOUT_2101010)

def SDL_ISPIXELFORMAT_FLOAT(format):
    return not SDL_ISPIXELFORMAT_FOURCC(format) and (SDL_PIXELTYPE(format) == SDL_PIXELTYPE_ARRAYF16 or SDL_PIXELTYPE(format) == SDL_PIXELTYPE_ARRAYF32)

def SDL_ISPIXELFORMAT_ALPHA(format):
    return SDL_ISPIXELFORMAT_PACKED(format) and (SDL_PIXELORDER(format) in [SDL_PACKEDORDER_ARGB, SDL_PACKEDORDER_RGBA, SDL_PACKEDORDER_ABGR, SDL_PACKEDORDER_BGRA]) \
        or SDL_ISPIXELFORMAT_ARRAY(format) and (SDL_PIXELORDER(format) in [SDL_ARRAYORDER_ARGB, SDL_ARRAYORDER_RGBA, SDL_ARRAYORDER_ABGR, SDL_ARRAYORDER_BGRA])

SDL_PixelFormat = ctypes.c_int

SDL_PIXELFORMAT_UNKNOWN = 0x00000000
SDL_PIXELFORMAT_INDEX1LSB = 0x11100100
SDL_PIXELFORMAT_INDEX1MSB = 0x11200100
SDL_PIXELFORMAT_INDEX2LSB = 0x1c100200
SDL_PIXELFORMAT_INDEX2MSB = 0x1c200200
SDL_PIXELFORMAT_INDEX4LSB = 0x12100400
SDL_PIXELFORMAT_INDEX4MSB = 0x12200400
SDL_PIXELFORMAT_INDEX8 = 0x13000801
SDL_PIXELFORMAT_RGB332 = 0x14110801
SDL_PIXELFORMAT_XRGB4444 = 0x15120c02
SDL_PIXELFORMAT_XBGR4444 = 0x15520c02
SDL_PIXELFORMAT_XRGB1555 = 0x15130f02
SDL_PIXELFORMAT_XBGR1555 = 0x15530f02
SDL_PIXELFORMAT_ARGB4444 = 0x15321002
SDL_PIXELFORMAT_RGBA4444 = 0x15421002
SDL_PIXELFORMAT_ABGR4444 = 0x15721002
SDL_PIXELFORMAT_BGRA4444 = 0x15821002
SDL_PIXELFORMAT_ARGB1555 = 0x15331002
SDL_PIXELFORMAT_RGBA5551 = 0x15441002
SDL_PIXELFORMAT_ABGR1555 = 0x15731002
SDL_PIXELFORMAT_BGRA5551 = 0x15841002
SDL_PIXELFORMAT_RGB565 = 0x15151002
SDL_PIXELFORMAT_BGR565 = 0x15551002
SDL_PIXELFORMAT_RGB24 = 0x17101803
SDL_PIXELFORMAT_BGR24 = 0x17401803
SDL_PIXELFORMAT_XRGB8888 = 0x16161804
SDL_PIXELFORMAT_RGBX8888 = 0x16261804
SDL_PIXELFORMAT_XBGR8888 = 0x16561804
SDL_PIXELFORMAT_BGRX8888 = 0x16661804
SDL_PIXELFORMAT_ARGB8888 = 0x16362004
SDL_PIXELFORMAT_RGBA8888 = 0x16462004
SDL_PIXELFORMAT_ABGR8888 = 0x16762004
SDL_PIXELFORMAT_BGRA8888 = 0x16862004
SDL_PIXELFORMAT_XRGB2101010 = 0x16172004
SDL_PIXELFORMAT_XBGR2101010 = 0x16572004
SDL_PIXELFORMAT_ARGB2101010 = 0x16372004
SDL_PIXELFORMAT_ABGR2101010 = 0x16772004
SDL_PIXELFORMAT_RGB48 = 0x18103006
SDL_PIXELFORMAT_BGR48 = 0x18403006
SDL_PIXELFORMAT_RGBA64 = 0x18204008
SDL_PIXELFORMAT_ARGB64 = 0x18304008
SDL_PIXELFORMAT_BGRA64 = 0x18504008
SDL_PIXELFORMAT_ABGR64 = 0x18604008
SDL_PIXELFORMAT_RGB48_FLOAT = 0x1a103006
SDL_PIXELFORMAT_BGR48_FLOAT = 0x1a403006
SDL_PIXELFORMAT_RGBA64_FLOAT = 0x1a204008
SDL_PIXELFORMAT_ARGB64_FLOAT = 0x1a304008
SDL_PIXELFORMAT_BGRA64_FLOAT = 0x1a504008
SDL_PIXELFORMAT_ABGR64_FLOAT = 0x1a604008
SDL_PIXELFORMAT_RGB96_FLOAT = 0x1b10600c
SDL_PIXELFORMAT_BGR96_FLOAT = 0x1b40600c
SDL_PIXELFORMAT_RGBA128_FLOAT = 0x1b208010
SDL_PIXELFORMAT_ARGB128_FLOAT = 0x1b308010
SDL_PIXELFORMAT_BGRA128_FLOAT = 0x1b508010
SDL_PIXELFORMAT_ABGR128_FLOAT = 0x1b608010
SDL_PIXELFORMAT_YV12 = 0x32315659
SDL_PIXELFORMAT_IYUV = 0x56555949
SDL_PIXELFORMAT_YUY2 = 0x32595559
SDL_PIXELFORMAT_UYVY = 0x59565955
SDL_PIXELFORMAT_YVYU = 0x55595659
SDL_PIXELFORMAT_NV12 = 0x3231564e
SDL_PIXELFORMAT_NV21 = 0x3132564e
SDL_PIXELFORMAT_P010 = 0x30313050
SDL_PIXELFORMAT_EXTERNAL_OES = 0x2053454f

if SDL_BYTEORDER == SDL_BIG_ENDIAN:
    SDL_PIXELFORMAT_RGBA32 = SDL_PIXELFORMAT_RGBA8888
    SDL_PIXELFORMAT_ARGB32 = SDL_PIXELFORMAT_ARGB8888
    SDL_PIXELFORMAT_BGRA32 = SDL_PIXELFORMAT_BGRA8888
    SDL_PIXELFORMAT_ABGR32 = SDL_PIXELFORMAT_ABGR8888
    SDL_PIXELFORMAT_RGBX32 = SDL_PIXELFORMAT_RGBX8888
    SDL_PIXELFORMAT_XRGB32 = SDL_PIXELFORMAT_XRGB8888
    SDL_PIXELFORMAT_BGRX32 = SDL_PIXELFORMAT_BGRX8888
    SDL_PIXELFORMAT_XBGR32 = SDL_PIXELFORMAT_XBGR8888

else:
    SDL_PIXELFORMAT_RGBA32 = SDL_PIXELFORMAT_ABGR8888
    SDL_PIXELFORMAT_ARGB32 = SDL_PIXELFORMAT_BGRA8888
    SDL_PIXELFORMAT_BGRA32 = SDL_PIXELFORMAT_ARGB8888
    SDL_PIXELFORMAT_ABGR32 = SDL_PIXELFORMAT_RGBA8888
    SDL_PIXELFORMAT_RGBX32 = SDL_PIXELFORMAT_XBGR8888
    SDL_PIXELFORMAT_XRGB32 = SDL_PIXELFORMAT_BGRX8888
    SDL_PIXELFORMAT_BGRX32 = SDL_PIXELFORMAT_XRGB8888
    SDL_PIXELFORMAT_XBGR32 = SDL_PIXELFORMAT_RGBX8888

SDL_ColorType = ctypes.c_int

SDL_COLOR_TYPE_UNKNOWN = 0
SDL_COLOR_TYPE_RGB = 1
SDL_COLOR_TYPE_YCBCR = 2

SDL_ColorRange = ctypes.c_int

SDL_COLOR_RANGE_UNKOWN = 0
SDL_COLOR_RANGE_LIMITED = 1
SDL_COLOR_RANGE_FULL = 2

SDL_ColorPrimaries = ctypes.c_int

SDL_COLOR_PRIMARIES_UNKNOWN = 0
SDL_COLOR_PRIMARIES_BT709 = 1
SDL_COLOR_PRIMARIES_UNSPECIFIED = 2
SDL_COLOR_PRIMARIES_BT470M = 4
SDL_COLOR_PRIMARIES_BT470BG = 5
SDL_COLOR_PRIMARIES_BT601 = 6
SDL_COLOR_PRIMARIES_SMPTE240 = 7
SDL_COLOR_PRIMARIES_GENERIC_FILM = 8
SDL_COLOR_PRIMARIES_BT2020 = 9
SDL_COLOR_PRIMARIES_XYZ = 10
SDL_COLOR_PRIMARIES_SMPTE431 = 11
SDL_COLOR_PRIMARIES_SMPTE432 = 12
SDL_COLOR_PRIMARIES_EBU3213 = 22
SDL_COLOR_PRIMARIES_CUSTOM = 31

SDL_TransferCharacteristics = ctypes.c_int

SDL_TRANSFER_CHARACTERISTICS_UNKNOWN = 0
SDL_TRANSFER_CHARACTERISTICS_BT709 = 1
SDL_TRANSFER_CHARACTERISTICS_UNSPECIFIED = 2
SDL_TRANSFER_CHARACTERISTICS_GAMMA22 = 4
SDL_TRANSFER_CHARACTERISTICS_GAMMA28 = 5
SDL_TRANSFER_CHARACTERISTICS_BT601 = 6
SDL_TRANSFER_CHARACTERISTICS_SMPTE240 = 7
SDL_TRANSFER_CHARACTERISTICS_LINEAR = 8
SDL_TRANSFER_CHARACTERISTICS_LOG100 = 9
SDL_TRANSFER_CHARACTERISTICS_LOG100_SQRT10 = 10
SDL_TRANSFER_CHARACTERISTICS_IEC61966 = 11
SDL_TRANSFER_CHARACTERISTICS_BT1361 = 12
SDL_TRANSFER_CHARACTERISTICS_SRGB = 13
SDL_TRANSFER_CHARACTERISTICS_BT2020_10BIT = 14
SDL_TRANSFER_CHARACTERISTICS_BT2020_12BIT = 15
SDL_TRANSFER_CHARACTERISTICS_PQ = 16
SDL_TRANSFER_CHARACTERISTICS_SMPTE428 = 17
SDL_TRANSFER_CHARACTERISTICS_HLG = 18
SDL_TRANSFER_CHARACTERISTICS_CUSTOM = 31

SDL_MatrixCoefficients = ctypes.c_int

SDL_MATRIX_COEFFICIENTS_IDENTITY = 0
SDL_MATRIX_COEFFICIENTS_BT709 = 1
SDL_MATRIX_COEFFICIENTS_UNSPECIFIED = 2
SDL_MATRIX_COEFFICIENTS_FCC = 4
SDL_MATRIX_COEFFICIENTS_BT470BG = 5
SDL_MATRIX_COEFFICIENTS_BT601 = 6
SDL_MATRIX_COEFFICIENTS_SMPTE240 = 7
SDL_MATRIX_COEFFICIENTS_YCGCO = 8
SDL_MATRIX_COEFFICIENTS_BT2020_NCL = 9
SDL_MATRIX_COEFFICIENTS_BT2020_CL = 10
SDL_MATRIX_COEFFICIENTS_SMPTE2085 = 11
SDL_MATRIX_COEFFICIENTS_CHROMA_DERIVED_NCL = 12
SDL_MATRIX_COEFFICIENTS_CHROMA_DERIVED_CL = 13
SDL_MATRIX_COEFFICIENTS_ICTCP = 14
SDL_MATRIX_COEFFICIENTS_CUSTOM = 31

SDL_ChromaLocation = ctypes.c_int

SDL_CHROMA_LOCATION_NONE = 0
SDL_CHROMA_LOCATION_LEFT = 1
SDL_CHROMA_LOCATION_CENTER = 2
SDL_CHROMA_LOCATION_TOPLEFT = 3

SDL_Colorspace = ctypes.c_int

SDL_COLORSPACE_UNKNOWN = 0x00000000
SDL_COLORSPACE_SRGB = 0x120005a0
SDL_COLORSPACE_SRGB_LINEAR = 0x12000500
SDL_COLORSPACE_HDR10 = 0x12002600
SDL_COLORSPACE_JPEG = 0x220004c6
SDL_COLORSPACE_BT601_LIMITED = 0x211018c6
SDL_COLORSPACE_BT601_FULL = 0x221018c6
SDL_COLORSPACE_BT709_LIMITED = 0x21100421
SDL_COLORSPACE_BT709_FULL = 0x22100421
SDL_COLORSPACE_BT2020_LIMITED = 0x21102609
SDL_COLORSPACE_BT2020_FULL = 0x22102609

SDL_COLORSPACE_RGB_DEFAULT = SDL_COLORSPACE_SRGB
SDL_COLORSPACE_YUV_DEFAULT = SDL_COLORSPACE_JPEG

class SDL_Color(ctypes.Structure):
    _fields_ = [
        ("r", ctypes.c_uint8),
        ("g", ctypes.c_uint8),
        ("b", ctypes.c_uint8),
        ("a", ctypes.c_uint8)
    ]

class SDL_FColor(ctypes.Structure):
    _fields_ = [
        ("r", ctypes.c_float),
        ("g", ctypes.c_float),
        ("b", ctypes.c_float),
        ("a", ctypes.c_float)
    ]

class SDL_Palette(ctypes.Structure):
    _fields_ = [
        ("ncolors", ctypes.c_int),
        ("colors", ctypes.POINTER(SDL_Color)),
        ("version", ctypes.c_uint32),
        ("refcount", ctypes.c_int)
    ]

class SDL_PixelFormatDetails(ctypes.Structure):
    _fields_ = [
        ("format", SDL_PixelFormat),
        ("bits_per_pixel", ctypes.c_uint8),
        ("bytes_per_pixel", ctypes.c_uint8),
        ("padding", ctypes.c_uint8 * 2),
        ("Rmask", ctypes.c_uint32),
        ("Gmask", ctypes.c_uint32),
        ("Bmask", ctypes.c_uint32),
        ("Amask", ctypes.c_uint32),
        ("Rbits", ctypes.c_uint8),
        ("Gbits", ctypes.c_uint8),
        ("Bbits", ctypes.c_uint8),
        ("Abits", ctypes.c_uint8),
        ("Rshift", ctypes.c_uint8),
        ("Gshift", ctypes.c_uint8),
        ("Bshift", ctypes.c_uint8),
        ("Ashift", ctypes.c_uint8),
        ("refcount", ctypes.c_int)
    ]

SDL_FUNC("SDL_GetPixelFormatName", ctypes.c_char_p, SDL_PixelFormat)
SDL_FUNC("SDL_GetMasksForPixelFormat", ctypes.c_bool, SDL_PixelFormat, ctypes.POINTER(ctypes.c_int32), ctypes.POINTER(ctypes.c_uint32), ctypes.POINTER(ctypes.c_uint32), ctypes.POINTER(ctypes.c_uint32), ctypes.POINTER(ctypes.c_uint32))
SDL_FUNC("SDL_GetPixelFormatForMasks", SDL_PixelFormat, ctypes.c_int32, ctypes.c_uint32, ctypes.c_uint32, ctypes.c_uint32, ctypes.c_uint32)
SDL_FUNC("SDL_GetPixelFormatDetails", ctypes.POINTER(SDL_PixelFormatDetails), SDL_PixelFormat)
SDL_FUNC("SDL_CreatePalette", ctypes.POINTER(SDL_Palette), ctypes.c_int)
SDL_FUNC("SDL_SetPaletteColors", ctypes.c_bool, ctypes.POINTER(SDL_Palette), ctypes.POINTER(SDL_Color), ctypes.c_int, ctypes.c_int)
SDL_FUNC("SDL_DestroyPalette", None, ctypes.POINTER(SDL_Palette))
SDL_FUNC("SDL_MapRGB", ctypes.c_uint32, ctypes.POINTER(SDL_PixelFormatDetails), ctypes.POINTER(SDL_Palette), ctypes.c_uint8, ctypes.c_uint8, ctypes.c_uint8)
SDL_FUNC("SDL_MapRGBA", ctypes.c_uint32, ctypes.POINTER(SDL_PixelFormatDetails), ctypes.POINTER(SDL_Palette), ctypes.c_uint8, ctypes.c_uint8, ctypes.c_uint8, ctypes.c_uint8)
SDL_FUNC("SDL_GetRGB", None, ctypes.c_uint32, ctypes.POINTER(SDL_PixelFormatDetails), ctypes.POINTER(SDL_Palette), ctypes.POINTER(ctypes.c_uint8), ctypes.POINTER(ctypes.c_uint8), ctypes.POINTER(ctypes.c_uint8))
SDL_FUNC("SDL_GetRGBA", None, ctypes.c_uint32, ctypes.POINTER(SDL_PixelFormatDetails), ctypes.POINTER(SDL_Palette), ctypes.POINTER(ctypes.c_uint8), ctypes.POINTER(ctypes.c_uint8), ctypes.POINTER(ctypes.c_uint8), ctypes.POINTER(ctypes.c_uint8))


from .__init__ import ctypes, \
    SDL_FUNC, SDL_SET_CURRENT_BINARY, SDL_BINARY

from.SDL_atomic import SDL_AtomicInt
from .SDL_thread import SDL_ThreadID

SDL_SET_CURRENT_BINARY(SDL_BINARY)

SDL_MUTEX_TIMEDOUT = 1

class SDL_Mutex(ctypes.c_void_p):
    ...

SDL_FUNC("SDL_CreateMutex", ctypes.POINTER(SDL_Mutex))
SDL_FUNC("SDL_LockMutex", None, ctypes.POINTER(SDL_Mutex))
SDL_FUNC("SDL_TryLockMutex", ctypes.c_bool, ctypes.POINTER(SDL_Mutex))
SDL_FUNC("SDL_UnlockMutex", None, ctypes.POINTER(SDL_Mutex))
SDL_FUNC("SDL_DestroyMutex", None, ctypes.POINTER(SDL_Mutex))

class SDL_RWLock(ctypes.c_void_p):
    ...

SDL_RWLOCK_TIMEDOUT = SDL_MUTEX_TIMEDOUT

SDL_FUNC("SDL_CreateRWLock", ctypes.POINTER(SDL_RWLock))
SDL_FUNC("SDL_LockRWLockForReading", None, ctypes.POINTER(SDL_RWLock))
SDL_FUNC("SDL_LockRWLockForWriting", None, ctypes.POINTER(SDL_RWLock))
SDL_FUNC("SDL_TryLockRWLockForReading", ctypes.c_bool, ctypes.POINTER(SDL_RWLock))
SDL_FUNC("SDL_TryLockRWLockForWriting", ctypes.c_bool, ctypes.POINTER(SDL_RWLock))
SDL_FUNC("SDL_UnlockRWLock", None, ctypes.POINTER(SDL_RWLock))
SDL_FUNC("SDL_DestroyRWLock", None, ctypes.POINTER(SDL_RWLock))

class SDL_Semaphore(ctypes.c_void_p):
    ...

SDL_FUNC("SDL_CreateSemaphore", ctypes.POINTER(SDL_Semaphore), ctypes.c_uint32)
SDL_FUNC("SDL_DestroySemaphore", None, ctypes.POINTER(SDL_Semaphore))
SDL_FUNC("SDL_WaitSemaphore", None, ctypes.POINTER(SDL_Semaphore))
SDL_FUNC("SDL_TryWaitSemaphore", ctypes.c_bool, ctypes.POINTER(SDL_Semaphore))
SDL_FUNC("SDL_WaitSemaphoreTimeout", ctypes.c_bool, ctypes.POINTER(SDL_Semaphore), ctypes.c_int32)
SDL_FUNC("SDL_SignalSemaphore", None, ctypes.POINTER(SDL_Semaphore))
SDL_FUNC("SDL_GetSemaphoreValue", ctypes.c_uint32, ctypes.POINTER(SDL_Semaphore))

class SDL_Condition(ctypes.c_void_p):
    ...

SDL_FUNC("SDL_CreateCondition", ctypes.POINTER(SDL_Condition))
SDL_FUNC("SDL_DestroyCondition", None, ctypes.POINTER(SDL_Condition))
SDL_FUNC("SDL_SignalCondition", None, ctypes.POINTER(SDL_Condition))
SDL_FUNC("SDL_BroadcastCondition", None, ctypes.POINTER(SDL_Condition))
SDL_FUNC("SDL_WaitCondition", None, ctypes.POINTER(SDL_Condition), ctypes.POINTER(SDL_Mutex))
SDL_FUNC("SDL_WaitConditionTimeout", ctypes.c_bool, ctypes.POINTER(SDL_Condition), ctypes.POINTER(SDL_Mutex), ctypes.c_int32)

SDL_InitStatus = ctypes.c_int

SDL_INIT_STATUS_UNINITIALIZED = 0
SDL_INIT_STATUS_INITIALIZING = 1
SDL_INIT_STATUS_INITIALIZED = 2
SDL_INIT_STATUS_UNINITIALIZING = 3

class SDL_InitState(ctypes.Structure):
    _fields_ = [
        ("status", SDL_AtomicInt),
        ("error", SDL_ThreadID),
        ("reserved", ctypes.c_void_p)
    ]

SDL_FUNC("SDL_ShouldInit", ctypes.c_bool, ctypes.POINTER(SDL_InitState))
SDL_FUNC("SDL_ShouldQuit", ctypes.c_bool, ctypes.POINTER(SDL_InitState))

SDL_FUNC("SDL_SetInitialized", None, ctypes.POINTER(SDL_InitState), ctypes.c_bool)


import sys, os, subprocess

try: __import__("embed").main(sys.argv)
except ModuleNotFoundError: ...

import sdl3, shader, ctypes, time, colorsys

def compileShader(name, size = ctypes.c_size_t()):
    subprocess.run(["glslc", f"{name}.glsl", "-o", f"{name}.spv"])
    data = sdl3.SDL_LoadFile(f"{name}.spv".encode(), ctypes.byref(size))
    return size.value, ctypes.cast(data, ctypes.POINTER(ctypes.c_uint8))

class Vertex(ctypes.Structure):
    _fields_ = [
        ("position", ctypes.c_float * 3),
        ("color", ctypes.c_float * 3)
    ]

class UniformData(ctypes.Structure):
    _fields_ = [
        ("color0", ctypes.c_float * 4),
        ("color1", ctypes.c_float * 4),
        ("color2", ctypes.c_float * 4)
    ]

@sdl3.SDL_main_func
def main(argc: ctypes.c_int, argv: sdl3.LP_c_char_p) -> ctypes.c_int:
    if not sdl3.SDL_Init(sdl3.SDL_INIT_VIDEO | sdl3.SDL_INIT_EVENTS | sdl3.SDL_INIT_AUDIO):
        print(f"failed to initialize library: {sdl3.SDL_GetError().decode().lower()}.")
        return -1
    
    window = sdl3.SDL_CreateWindow("Aermoss".encode(), 1200, 600, sdl3.SDL_WINDOW_RESIZABLE)
    event, running = sdl3.SDL_Event(), True

    gpuDrivers = [sdl3.SDL_GetGPUDriver(i).decode() for i in range(sdl3.SDL_GetNumGPUDrivers())]
    tryGetDriver, tryUseVulkan = lambda order, drivers: next((i for i in order if i in drivers), None), True
    gpuDriver = tryGetDriver(["vulkan"] if tryUseVulkan else [], gpuDrivers)
    print(f"available gpu drivers: {', '.join(gpuDrivers)}. (current: {gpuDriver})")
    device = sdl3.SDL_CreateGPUDevice(sdl3.SDL_GPU_SHADERFORMAT_SPIRV, True, gpuDriver.encode())

    if not device:
        print(f"failed to create gpu device: {sdl3.SDL_GetError().decode().lower()}.")
        return -1

    if not sdl3.SDL_ClaimWindowForGPUDevice(device, window):
        print(f"failed to claim window for gpu device: {sdl3.SDL_GetError().decode().lower()}")
        return -1

    vertexShader = sdl3.SDL_CreateGPUShader(device, sdl3.SDL_GPUShaderCreateInfo(*shader.vertexData, entrypoint = "main".encode(), format = sdl3.SDL_GPU_SHADERFORMAT_SPIRV, stage = sdl3.SDL_GPU_SHADERSTAGE_VERTEX, num_uniform_buffers = 1))
    fragmentShader = sdl3.SDL_CreateGPUShader(device, sdl3.SDL_GPUShaderCreateInfo(*shader.fragmentData, entrypoint = "main".encode(), format = sdl3.SDL_GPU_SHADERFORMAT_SPIRV, stage = sdl3.SDL_GPU_SHADERSTAGE_FRAGMENT))

    pipeline = sdl3.SDL_CreateGPUGraphicsPipeline(device, sdl3.SDL_GPUGraphicsPipelineCreateInfo(vertexShader, fragmentShader, primitive_type = sdl3.SDL_GPU_PRIMITIVETYPE_TRIANGLELIST,
        target_info = sdl3.SDL_GPUGraphicsPipelineTargetInfo(*sdl3.SDL_ARRAY(sdl3.SDL_GPUColorTargetDescription(sdl3.SDL_GetGPUSwapchainTextureFormat(device, window)))),
        vertex_input_state = sdl3.SDL_GPUVertexInputState(*sdl3.SDL_ARRAY(sdl3.SDL_GPUVertexBufferDescription(0, ctypes.sizeof(Vertex), sdl3.SDL_GPU_VERTEXINPUTRATE_VERTEX, 0)),
            *sdl3.SDL_ARRAY(sdl3.SDL_GPUVertexAttribute(0, 0, sdl3.SDL_GPU_VERTEXELEMENTFORMAT_FLOAT3, 0)))
    ))
    
    sdl3.SDL_ReleaseGPUShader(device, fragmentShader)
    sdl3.SDL_ReleaseGPUShader(device, vertexShader)
    
    if not pipeline:
        print(f"failed to create gpu pipeline: {sdl3.SDL_GetError().decode().lower()}.")
        return -1
    
    vertexBuffer = sdl3.SDL_CreateGPUBuffer(device, sdl3.SDL_GPUBufferCreateInfo(sdl3.SDL_GPU_BUFFERUSAGE_VERTEX, ctypes.sizeof(Vertex) * 3))
    transferBuffer = sdl3.SDL_CreateGPUTransferBuffer(device, sdl3.SDL_GPUTransferBufferCreateInfo(sdl3.SDL_GPU_TRANSFERBUFFERUSAGE_UPLOAD, ctypes.sizeof(Vertex) * 3))

    transferData = ctypes.cast(sdl3.SDL_MapGPUTransferBuffer(device, transferBuffer, False), ctypes.POINTER(Vertex))
    transferData[0] = Vertex(sdl3.SDL_ARRAY(-0.5, -0.5, 0.0, type = ctypes.c_float)[0])
    transferData[1] = Vertex(sdl3.SDL_ARRAY( 0.5, -0.5, 0.0, type = ctypes.c_float)[0])
    transferData[2] = Vertex(sdl3.SDL_ARRAY( 0.0,  0.5, 0.0, type = ctypes.c_float)[0])
    sdl3.SDL_UnmapGPUTransferBuffer(device, transferBuffer)

    commandBuffer = sdl3.SDL_AcquireGPUCommandBuffer(device)
    copyPass = sdl3.SDL_BeginGPUCopyPass(commandBuffer)
    sdl3.SDL_UploadToGPUBuffer(copyPass, sdl3.SDL_GPUTransferBufferLocation(transferBuffer, 0), sdl3.SDL_GPUBufferRegion(vertexBuffer, 0, ctypes.sizeof(Vertex) * 3), False)
    sdl3.SDL_EndGPUCopyPass(copyPass)

    sdl3.SDL_SubmitGPUCommandBuffer(commandBuffer)
    sdl3.SDL_ReleaseGPUTransferBuffer(device, transferBuffer)
    uniformData, lastTime, hue = UniformData(), time.time(), 0.0

    while running:
        while sdl3.SDL_PollEvent(ctypes.byref(event)):
            match event.type:
                case sdl3.SDL_EVENT_QUIT:
                    running = False

                case sdl3.SDL_EVENT_KEY_DOWN:
                    if event.key.key == sdl3.SDLK_ESCAPE:
                        running = False

        commandBuffer = sdl3.SDL_AcquireGPUCommandBuffer(device)

        if not commandBuffer:
            print(f"failed to acquire gpu command buffer: {sdl3.SDL_GetError().decode().lower()}")
            return -1

        swapChainTexture = ctypes.POINTER(sdl3.SDL_GPUTexture)()
        sdl3.SDL_WaitAndAcquireGPUSwapchainTexture(commandBuffer, window, ctypes.byref(swapChainTexture), None, None)

        if not swapChainTexture:
            print(f"failed to acquire gpu swapchain texture: {sdl3.SDL_GetError().decode().lower()}")
            return -1
        
        lastTime, deltaTime = \
            time.time(), time.time() - lastTime

        colorTargetInfo = sdl3.SDL_GPUColorTargetInfo(swapChainTexture, load_op = sdl3.SDL_GPU_LOADOP_CLEAR,
            clear_color = sdl3.SDL_FColor(*colorsys.hsv_to_rgb(hue := (hue + 0.25 * deltaTime), 1.0, 0.0), 1.0))
        
        uniformData.color0 = sdl3.SDL_ARRAY(*colorsys.hsv_to_rgb(hue + 0.48, 1.0, 1.0), 1.0, type = ctypes.c_float)[0]
        uniformData.color1 = sdl3.SDL_ARRAY(*colorsys.hsv_to_rgb(hue + 0.32, 1.0, 1.0), 1.0, type = ctypes.c_float)[0]
        uniformData.color2 = sdl3.SDL_ARRAY(*colorsys.hsv_to_rgb(hue + 0.64, 1.0, 1.0), 1.0, type = ctypes.c_float)[0]

        renderPass = sdl3.SDL_BeginGPURenderPass(commandBuffer, ctypes.byref(colorTargetInfo), 1, None)
        sdl3.SDL_BindGPUGraphicsPipeline(renderPass, pipeline)
        sdl3.SDL_BindGPUVertexBuffers(renderPass, 0, sdl3.SDL_GPUBufferBinding(vertexBuffer, 0), 1)
        sdl3.SDL_PushGPUVertexUniformData(commandBuffer, 0, ctypes.byref(uniformData), ctypes.sizeof(uniformData))
        sdl3.SDL_DrawGPUPrimitives(renderPass, 3, 1, 0, 0)
        sdl3.SDL_EndGPURenderPass(renderPass)

        sdl3.SDL_SubmitGPUCommandBuffer(commandBuffer)

    sdl3.SDL_ReleaseGPUBuffer(device, vertexBuffer)
    sdl3.SDL_ReleaseGPUGraphicsPipeline(device, pipeline)
    sdl3.SDL_DestroyGPUDevice(device)

    sdl3.SDL_DestroyWindow(window)
    sdl3.SDL_Quit()
    return 0


import os, sdl3, ctypes, colorsys, time

@sdl3.SDL_main_func
def main(argc: ctypes.c_int, argv: sdl3.LP_c_char_p) -> ctypes.c_int:
    print(f"total lines of code: {sum([len(open(f'sdl3/{i}', 'r').readlines()) for i in os.listdir('sdl3') if i.endswith('.py')])}.")
    print(f"loaded {sum(len(v) for k, v in sdl3.modules.items())} functions.")

    if not sdl3.SDL_Init(sdl3.SDL_INIT_VIDEO | sdl3.SDL_INIT_EVENTS | sdl3.SDL_INIT_AUDIO):
        print(f"failed to initialize library: {sdl3.SDL_GetError().decode().lower()}.")
        return -1
    
    window = sdl3.SDL_CreateWindow("Aermoss".encode(), 1200, 600, sdl3.SDL_WINDOW_RESIZABLE)

    renderDrivers = [sdl3.SDL_GetRenderDriver(i).decode() for i in range(sdl3.SDL_GetNumRenderDrivers())]
    tryGetDriver, tryUseVulkan = lambda order, drivers: next((i for i in order if i in drivers), None), False
    renderDriver = tryGetDriver((["vulkan"] if tryUseVulkan else []) + ["opengl", "software"], renderDrivers)
    print(f"available render drivers: {', '.join(renderDrivers)} (current: {renderDriver}).")
    renderer = sdl3.SDL_CreateRenderer(window, renderDriver.encode())

    if not renderer:
        print(f"failed to create renderer: {sdl3.SDL_GetError().decode().lower()}.")
        return -1
    
    audioDrivers = [sdl3.SDL_GetAudioDriver(i).decode() for i in range(sdl3.SDL_GetNumAudioDrivers())]
    print(f"available audio drivers: {', '.join(audioDrivers)} (current: {sdl3.SDL_GetCurrentAudioDriver().decode().lower()}).")
    audioDevices = sdl3.SDL_GetAudioPlaybackDevices(None)

    if not audioDevices:
        print(f"failed to get audio devices: {sdl3.SDL_GetError().decode().lower()}.")
        return -1

    audioDevice = sdl3.SDL_AUDIO_DEVICE_DEFAULT_PLAYBACK if True else audioDevices[0]
    currentAudioDevice = sdl3.SDL_OpenAudioDevice(audioDevice, None)
    
    if currentAudioDevice:
        print(f"current audio device: {sdl3.SDL_GetAudioDeviceName(currentAudioDevice).decode().lower()}.")

    surface = sdl3.IMG_Load("res/example.png".encode())
    texture = sdl3.SDL_CreateTextureFromSurface(renderer, surface)

    rect = sdl3.SDL_Rect()
    sdl3.SDL_GetSurfaceClipRect(surface, ctypes.byref(rect))

    frect = sdl3.SDL_FRect()
    sdl3.SDL_RectToFRect(ctypes.byref(rect), ctypes.byref(frect))
    running, hue, lastTime, scale = True, 0.0, time.time(), 0.75

    sdl3.TTF_Init()
    font = sdl3.TTF_OpenFont("res/example.ttf".encode(), 32.0)

    if currentAudioDevice:
        sdl3.Mix_Init(sdl3.MIX_INIT_WAVPACK)
        audioSpec = sdl3.SDL_AudioSpec()
        sdl3.Mix_OpenAudio(currentAudioDevice, ctypes.byref(audioSpec))
        chunks = [sdl3.Mix_LoadWAV(f"res/voice/{i}".encode()) for i in os.listdir("res/voice")]
        currentIndex, channel = 0, 0

    frames, frameCooldown = 0.0, 1.0
    textTexture, textFRect = None, sdl3.SDL_FRect()
    sinceLastFrame = frameCooldown

    event = sdl3.SDL_Event()

    while running:
        while sdl3.SDL_PollEvent(ctypes.byref(event)):
            match event.type:
                case sdl3.SDL_EVENT_QUIT:
                    running = False

                case sdl3.SDL_EVENT_KEY_DOWN:
                    if event.key.key == sdl3.SDLK_ESCAPE:
                        running = False

        width, height = ctypes.c_int(), ctypes.c_int()
        sdl3.SDL_GetWindowSize(window, width, height)

        frect.w, frect.h = frect.w * width.value / frect.w * scale, frect.h * width.value / frect.w * scale
        frect.x, frect.y = width.value / 2 - frect.w / 2, height.value / 2 - frect.h / 2

        if currentAudioDevice and not sdl3.Mix_Playing(channel):
            channel = sdl3.Mix_PlayChannel(-1, chunks[currentIndex], 1)
            currentIndex += 1

            if currentIndex >= len(chunks):
                currentIndex = 0

        lastTime, deltaTime = \
            time.time(), time.time() - lastTime

        hue, frames = (hue + 0.5 * deltaTime) % 1.0, frames + 1.0
        sdl3.SDL_SetRenderDrawColorFloat(renderer, *colorsys.hsv_to_rgb(hue, 1.0, 0.1), 1.0)
        sdl3.SDL_RenderClear(renderer)
        sdl3.SDL_RenderTexture(renderer, texture, None, ctypes.byref(frect))
        sinceLastFrame += deltaTime

        if sinceLastFrame >= frameCooldown:
            framesPerSecond = int(frames / sinceLastFrame)
            sinceLastFrame, frames = 0.0, 0.0

            if textTexture is not None:
                sdl3.SDL_DestroySurface(textSurface)
                sdl3.SDL_DestroyTexture(textTexture)

            textSurface = sdl3.TTF_RenderText_Blended(font, f"FPS: {framesPerSecond}".encode(), 0, sdl3.SDL_Color(255, 255, 255, 255))
            textTexture = sdl3.SDL_CreateTextureFromSurface(renderer, textSurface)

            textRect = sdl3.SDL_Rect()
            sdl3.SDL_GetSurfaceClipRect(textSurface, ctypes.byref(textRect))
            sdl3.SDL_RectToFRect(ctypes.byref(textRect), ctypes.byref(textFRect))

        error = sdl3.SDL_GetError()

        if error:
            print(f"error: {error.decode().lower()}.")
            return -1

        if textTexture is not None:
            sdl3.SDL_RenderTexture(renderer, textTexture, None, ctypes.byref(textFRect))

        sdl3.SDL_RenderPresent(renderer)

    if textTexture is not None:
        sdl3.SDL_DestroySurface(textSurface)
        sdl3.SDL_DestroyTexture(textTexture)

    if currentAudioDevice:
        for i in chunks:
            sdl3.Mix_FreeChunk(i)

        sdl3.Mix_CloseAudio()
        sdl3.Mix_Quit()

    sdl3.TTF_CloseFont(font)
    sdl3.TTF_Quit()

    sdl3.SDL_DestroySurface(surface)
    sdl3.SDL_DestroyTexture(texture)

    sdl3.SDL_DestroyRenderer(renderer)
    sdl3.SDL_DestroyWindow(window)
    sdl3.SDL_Quit()
    return 0